# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
Contains visitors to lower GTIR to DaCe SDFG.

Note: this module covers the fieldview flavour of GTIR.
"""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Dict, Iterable, List, Mapping, Optional, Protocol, Sequence, Tuple, Union

import dace
from dace import nodes as dace_nodes, subsets as dace_subsets
from dace.frontend.python import astutils as dace_astutils

from gt4py import eve
from gt4py.eve import concepts
from gt4py.next import common as gtx_common, config as gtx_config, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)
from gt4py.next.iterator.transforms import prune_casts as ir_prune_casts, symbol_ref_utils
from gt4py.next.iterator.type_system import inference as gtir_type_inference
from gt4py.next.program_processors.runners.dace import sdfg_args as gtx_dace_args
from gt4py.next.program_processors.runners.dace.lowering import (
    gtir_domain,
    gtir_to_sdfg_concat_where,
    gtir_to_sdfg_primitives,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import type_specifications as ts, type_translation as tt


def _replace_connectors_in_code_string(
    code: str, language: dace.dtypes.Language, connector_mapping: Mapping[str, str]
) -> str:
    """Helper function to replace connector names in the code of a Python tasklet."""
    code_block = dace.properties.CodeBlock(code, language)
    transformed_code_stmts = [
        dace_astutils.ASTFindReplace(connector_mapping).visit(stmt) for stmt in code_block.code
    ]
    return dace.properties.CodeBlock(transformed_code_stmts, language).as_string


class DataflowBuilder(Protocol):
    """Visitor interface to build a dataflow subgraph."""

    @abc.abstractmethod
    def get_offset_provider_type(self, offset: str) -> gtx_common.OffsetProviderTypeElem: ...

    @abc.abstractmethod
    def unique_nsdfg_name(self, prefix: str) -> str: ...

    @abc.abstractmethod
    def unique_map_name(self, name: str) -> str: ...

    @abc.abstractmethod
    def unique_tasklet_name(self, name: str) -> str: ...

    @abc.abstractmethod
    def unique_temp_name(self) -> str: ...

    def add_temp_array(
        self, sdfg: dace.SDFG, shape: Sequence[Any], dtype: dace.dtypes.typeclass
    ) -> tuple[str, dace.data.Scalar]:
        """Add a temporary array to the SDFG."""
        temp_name = self.unique_temp_name()
        return sdfg.add_transient(temp_name, shape, dtype)

    def add_temp_array_like(
        self, sdfg: dace.SDFG, datadesc: dace.data.Array
    ) -> tuple[str, dace.data.Scalar]:
        """Add a temporary array to the SDFG."""
        temp_name = self.unique_temp_name()
        return sdfg.add_temp_transient_like(datadesc, name=temp_name)

    def add_temp_scalar(
        self, sdfg: dace.SDFG, dtype: dace.dtypes.typeclass
    ) -> tuple[str, dace.data.Scalar]:
        """Add a temporary scalar to the SDFG."""
        temp_name = self.unique_temp_name()
        return sdfg.add_scalar(temp_name, dtype, transient=True)

    def add_map(
        self,
        name: str,
        state: dace.SDFGState,
        ndrange: Union[
            Dict[str, Union[str, dace.subsets.Subset]],
            List[Tuple[str, Union[str, dace.subsets.Subset]]],
        ],
        **kwargs: Any,
    ) -> Tuple[dace_nodes.MapEntry, dace_nodes.MapExit]:
        """Wrapper of `dace.SDFGState.add_map` that assigns unique name."""
        unique_name = self.unique_map_name(name)
        return state.add_map(unique_name, ndrange, **kwargs)

    def add_tasklet(
        self,
        name: str,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        inputs: set[str] | Mapping[str, dace.dtypes.typeclass | None],
        outputs: set[str] | Mapping[str, dace.dtypes.typeclass | None],
        code: str,
        language: dace.dtypes.Language = dace.dtypes.Language.Python,
        **kwargs: Any,
    ) -> dace_nodes.Tasklet:
        """Wrapper of `dace.SDFGState.add_tasklet` that assigns a unique name.

        It also modifies the tasklet connectors by adding a prefix string (see
        `gtir_to_sdfg_utils.get_tasklet_connector()`), in order to avoid name conflicts
        with SDFG data. Otherwise, SDFG validation would detect such conflicts and fail.
        """
        if isinstance(inputs, set):
            inputs = {k: None for k in sorted(inputs)}
        if isinstance(outputs, set):
            outputs = {k: None for k in sorted(outputs)}
        assert inputs.keys().isdisjoint(outputs.keys())

        connector_mapping = {
            conn: gtir_to_sdfg_utils.make_tasklet_connector_for(conn)
            for conn in (inputs.keys() | outputs.keys())
        }
        new_code = _replace_connectors_in_code_string(code, language, connector_mapping)

        inputs = {connector_mapping[k]: v for k, v in inputs.items()}
        outputs = {connector_mapping[k]: v for k, v in outputs.items()}

        unique_name = self.unique_tasklet_name(name)
        tasklet_node = state.add_tasklet(
            unique_name, inputs, outputs, new_code, language=language, **kwargs
        )
        return tasklet_node, connector_mapping

    def add_mapped_tasklet(
        self,
        name: str,
        sdfg: dace.SDFG,
        state: dace.SDFGState,
        map_ranges: Mapping[str, str | dace.subsets.Subset],
        inputs: Mapping[str, dace.Memlet],
        code: str,
        outputs: Mapping[str, dace.Memlet],
        language: dace.dtypes.Language = dace.dtypes.Language.Python,
        **kwargs: Any,
    ) -> tuple[dace_nodes.Tasklet, dace_nodes.MapEntry, dace_nodes.MapExit, dict[str, str]]:
        """Wrapper of `dace.SDFGState.add_mapped_tasklet` that assigns a unique name.

        It also modifies the tasklet connectors, in the same way as `add_tasklet()`.
        """
        assert inputs.keys().isdisjoint(outputs.keys())

        connector_mapping = {
            conn: gtir_to_sdfg_utils.make_tasklet_connector_for(conn)
            for conn in (inputs.keys() | outputs.keys())
        }
        new_code = _replace_connectors_in_code_string(code, language, connector_mapping)

        inputs = {connector_mapping[k]: v for k, v in inputs.items()}
        outputs = {connector_mapping[k]: v for k, v in outputs.items()}
        unique_name = self.unique_tasklet_name(name)
        tasklet_node, map_entry, map_exit = state.add_mapped_tasklet(
            unique_name, map_ranges, inputs, new_code, outputs, language=language, **kwargs
        )
        return tasklet_node, map_entry, map_exit, connector_mapping


@dataclasses.dataclass(frozen=True)
class SubgraphContext:
    """Represents the subgraph context in which to lower a GTIR expression to dataflow."""

    sdfg: dace.SDFG
    state: dace.SDFGState
    scope_symbols: dict[str, ts.DataType]

    def get_symbol_type(self, symbol_name: str) -> ts.DataType:
        return self.scope_symbols[symbol_name]

    def get_input_data(self) -> set[str]:
        """Retrieve the names of arrays and scalars which are input data to this context.

        Note that only the data that is currently used in this context, in other words
        accessed in this SDFG state, is included in the returned set.
        """

        flat_symbols = []
        for sym_name, sym_type in self.scope_symbols.items():
            if isinstance(sym_type, ts.TupleType):
                flat_symbols.extend(gtir_to_sdfg_utils.flatten_tuple_fields(sym_name, sym_type))
            else:
                flat_symbols.append(im.sym(sym_name, sym_type))

        used_data = {access_node.data for access_node in self.state.data_nodes()}
        input_data = {data_name for sym in flat_symbols if (data_name := str(sym.id)) in used_data}
        assert not any(self.sdfg.arrays[data_name].transient for data_name in input_data)
        return input_data

    def copy_data(
        self,
        sdfg_builder: SDFGBuilder,
        src: gtir_to_sdfg_types.FieldopData,
        domain: gtir_domain.FieldopDomain | None,
    ) -> gtir_to_sdfg_types.FieldopData:
        """Copy data from an access node into a new data buffer in this SDFG context.

        If 'domain=None', it allocates transient scalar or array storage with
        the exact same descriptor as the 'src' data in current SDFG context, and
        creates an edge memlet to write the full shape.
        If the 'domain' passed as argument is not None, the result field is defined
        on the given domain, and only the corresponding subset is copied.

        Args:
            src: The data access node to copy, either a scalar or array.
            domain: If None, the full shape is copied. Otherwise, the given domain
                is translated into the subset to be copied.

        Returns:
            The new access node where the data is copied.
        """
        data_desc = src.dc_node.desc(self.sdfg)
        if isinstance(src.gt_type, ts.FieldType):
            if domain is None:
                out, out_desc = sdfg_builder.add_temp_array_like(self.sdfg, data_desc)
                out_origin = list(src.origin)
                src_subset = ",".join(f"0:{size}" for size in data_desc.shape)
            else:
                out_dims, out_origin, out_shape = gtir_domain.get_field_layout(domain)
                assert out_dims == src.gt_type.dims
                out, out_desc = sdfg_builder.add_temp_array(self.sdfg, out_shape, data_desc.dtype)
                src_subset = ",".join(
                    f"{dst_origin - src_origin}:{dst_origin - src_origin + size}"
                    for dst_origin, src_origin, size in zip(
                        out_origin, src.origin, out_shape, strict=True
                    )
                )
        else:
            assert domain is None
            assert isinstance(data_desc, dace.data.Scalar)
            out, out_desc = sdfg_builder.add_temp_array_like(self.sdfg, data_desc)
            out_origin = []
            src_subset = "0"

        out_node = self.state.add_access(out)
        out_subset = dace_subsets.Range.from_array(out_desc)
        self.state.add_nedge(
            src.dc_node,
            out_node,
            dace.Memlet(out, subset=out_subset, other_subset=src_subset),
        )
        return gtir_to_sdfg_types.FieldopData(out_node, src.gt_type, tuple(out_origin))

    def map_nsdfg_field(
        self,
        sdfg_builder: SDFGBuilder,
        nsdfg_field: gtir_to_sdfg_types.FieldopData,
        nsdfg: dace.SDFG,
        symbol_mapping: dict[str, dace.symbolic.SymbolicType],
    ) -> gtir_to_sdfg_types.FieldopData:
        """
        Make the data descriptor which 'nsdfg_field' refers to, and which is located
        inside a nested SDFG, available in this context.

        This means to turn 'nsdfg_field' into a global array and create a new data
        descriptor inside this SDFG context, with same shape and strides.

        Args:
            sdfg_builder: The object used to build the SDFG in this context.
            nsdfg_field: The descriptor of the field inside the nested SDFG.
            nsdfg: The nested SDFG where 'nsdfg_field' is defined.
            symbol_mapping: Mapping of symbols from the nested SDFG to the SDFG
                in this context.

        Returns:
            The descriptor of the new field inside this SDFG context.
        """
        inner_desc = nsdfg_field.dc_node.desc(nsdfg)
        assert inner_desc.transient
        inner_desc.transient = False

        if isinstance(nsdfg_field.gt_type, ts.ScalarType):
            outer, outer_desc = sdfg_builder.add_temp_scalar(self.sdfg, inner_desc.dtype)
            outer_origin = []
        else:
            outer, outer_desc = sdfg_builder.add_temp_array_like(self.sdfg, inner_desc)
            # We cannot use a copy of the inner data descriptor directly, we have to apply the symbol mapping.
            dace.symbolic.safe_replace(
                symbol_mapping,
                lambda m: dace.sdfg.replace_properties_dict(outer_desc, m),
            )
            # Same applies to the symbols used as field origin (the domain range start)
            outer_origin = [
                gtir_to_sdfg_utils.safe_replace_symbolic(val, symbol_mapping)
                for val in nsdfg_field.origin
            ]

        outer_node = self.state.add_access(outer)
        return gtir_to_sdfg_types.FieldopData(outer_node, nsdfg_field.gt_type, tuple(outer_origin))


class SDFGBuilder(DataflowBuilder, Protocol):
    """Visitor interface available to GTIR-primitive translators."""

    @abc.abstractmethod
    def make_field(
        self,
        data_node: dace_nodes.AccessNode,
        data_type: ts.FieldType,
    ) -> gtir_to_sdfg_types.FieldopData:
        """Retrieve the field descriptor of a data node, including the origin information.

        In case of `ScalarType` data, the `FieldopData` is constructed with `origin=None`.
        In case of `FieldType` data, the field origin is added to the data descriptor.
        Besides, if the `FieldType` contains a local dimension, the descriptor is converted
        to a canonical form where the field domain consists of all global dimensions
        (the grid axes) and the field data type is `ListType`, with `offset_type` equal
        to the field local dimension.

        TODO(edoapo): consider refactoring this method and moving it to a type module
            close to the `FieldopData` type declaration.

        Args:
            data_node: The access node to the SDFG data storage.
            data_type: The GT4Py data descriptor, which can either come from a program
                symbol, or from an intermediate field for an argument expression.

        Returns:
            The descriptor associated with the SDFG data storage, filled with field origin.
        """
        ...

    @abc.abstractmethod
    def is_column_axis(self, dim: gtx_common.Dimension) -> bool:
        """Check if the given dimension is the column axis."""
        ...

    @abc.abstractmethod
    def setup_nested_context(
        self,
        expr: gtir.Lambda,
        sdfg_name: str,
        parent_ctx: SubgraphContext,
        params: Iterable[gtir.Sym],
        symbolic_inputs: set[str],
        capture_scope_symbols: bool,
    ) -> SubgraphContext:
        """
        Create an nested SDFG context to lower a lambda expression, indipendent
        from the current context where the parent expression is being translated.

        This method will setup the global symbols, that correspond to the parameters
        of the expression to be lowered, as well as the set of symbolic arguments,
        that is symbols used in domain expressions or scalar values represented
        as dace symbols in the parent SDFG.

        Args:
            expr: The lambda expression to be lowered as a nested SDFG.
            sdfg_name: The name of the nested SDFG where to lower the given expression.
            parent_ctx: The parent SDFG context.
            params: List of GTIR symbols passed as parameters to the lambda expression.
            symbolic_inputs: Arguments that have to be passed to the nested SDFG
                as dace symbols.
            capture_scope_symbols: When True, the lambda expression will capture
                GTIR symbols defined in the parent scope.

        Returns:
            The new SDFG context in which to lower a nested expression.
        """
        ...

    @abc.abstractmethod
    def add_nested_sdfg(
        self,
        node: gtir.Lambda,
        inner_ctx: SubgraphContext,
        outer_ctx: SubgraphContext,
        symbolic_args: Mapping[str, gtir_to_sdfg_types.SymbolicData],
        data_args: Mapping[str, gtir_to_sdfg_types.FieldopData | None],
        inner_result: gtir_to_sdfg_types.FieldopResult,
        capture_outer_data: bool,
    ) -> tuple[dace_nodes.NestedSDFG, Mapping[str, dace.Memlet]]:
        """
        Helper function that prepares the input connections and symbol mapping before
        calling `SDFG.add_nestd_sdfg()` to add the given SDFG as a nested SDFG node
        inside the parent SDFG.

        Args:
            node: The lambda GTIR node containing the expression which was lowered in `inner_ctx`.
            inner_ctx: The nested SDFG context, containing the state where `inner_result` is written.
            outer_ctx: The parent SDFG context, containing the state where the nested SDFG should be added.
            symbolic_args: Scalar argumemts to be passed to the nested SDFG through symbol mapping.
            data_args: Data arguments to be passed through edge memlets. It contains `None` values
                for those arguments that, based on domain inference, should not be used.
            inner_result: The data produced by the nested SDFG, inside the state specified by `inner_ctx`.
            capture_outer_data: Allow capturing scalars and arrays defined in the parent SDFG.

        Returns:
            A tuple of two elements:
            - The nested SDFG graph node.
            - The mapping from input connectors to data memlets.
        """
        ...

    @abc.abstractmethod
    def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        """Visit a node of the GT4Py IR."""
        ...


def _flatten_tuple_symbols(symbols: Iterable[gtir.Sym]) -> list[gtir.Sym]:
    """
    Helper function to flatten tuple symbols, recursively in case of nested tuples,
    and extract all scalar symbols.
    """
    flat_symbols: list[gtir.Sym] = []
    for sym in symbols:
        if isinstance(sym.type, ts.TupleType):
            flat_symbols.extend(
                f for f in gtir_to_sdfg_utils.flatten_tuple_fields(sym.id, sym.type)
            )
        else:
            flat_symbols.append(sym)
    return flat_symbols


def _make_access_index_for_field(
    domain: gtir_domain.FieldopDomain, data: gtir_to_sdfg_types.FieldopData
) -> dace.subsets.Range:
    """Helper method to build a memlet subset of a field over the given domain."""
    # convert domain expression to dictionary to ease access to the dimensions,
    # since the access indices have to follow the order of dimensions in field domain
    if isinstance(data.gt_type, ts.FieldType) and len(data.gt_type.dims) != 0:
        assert data.origin is not None
        domain_ranges = {
            domain_range.dim: (domain_range.start, domain_range.stop) for domain_range in domain
        }
        return dace.subsets.Range(
            (domain_ranges[dim][0] - origin, domain_ranges[dim][1] - origin - 1, 1)
            for dim, origin in zip(data.gt_type.dims, data.origin, strict=True)
        )
    else:
        assert len(domain) == 0
        return dace.subsets.Range.from_string("0")


def flatten_tuple_args(
    args: Iterable[
        tuple[gtir.Sym, gtir_to_sdfg_types.FieldopResult | gtir_to_sdfg_types.SymbolicData]
    ],
) -> tuple[
    dict[str, gtir_to_sdfg_types.FieldopData | None],
    dict[str, gtir_to_sdfg_types.SymbolicData],
]:
    """Helper function to flatten tuple arguments passed to a lambda node.

    Args:
        args: A list of arguments containing either a symbolic value or some data
            access node, possibly in the form of tuple.

    Return:
        Two dictionaries of flat arguments, including all nested fields extracted
        from tuples, one for data arguments and one for symbolic arguments.
    """
    data_args: dict[str, gtir_to_sdfg_types.FieldopData | None] = {}
    symb_args: dict[str, gtir_to_sdfg_types.SymbolicData] = {}
    for gtsym, arg in args:
        gtsym_id = str(gtsym.id)
        if isinstance(arg, tuple):
            assert isinstance(gtsym.type, ts.TupleType)
            tuple_symbols = gtir_to_sdfg_utils.flatten_tuple_fields(gtsym_id, gtsym.type)
            tuple_args = gtx_utils.flatten_nested_tuple(arg)
            tuple_data_args, tuple_symb_args = flatten_tuple_args(
                zip(tuple_symbols, tuple_args, strict=True)
            )
            data_args |= tuple_data_args
            symb_args |= tuple_symb_args
        elif isinstance(arg, gtir_to_sdfg_types.SymbolicData):
            symb_args[gtsym_id] = arg
        else:
            data_args[gtsym_id] = arg

    return data_args, symb_args


@dataclasses.dataclass(frozen=True)
class GTIRToSDFG(eve.NodeVisitor, SDFGBuilder):
    """Provides translation capability from a GTIR program to a DaCe SDFG.

    A single instance of this visitor is used for the entire lowering, across all
    levels of nested SDFGs. The level-specific information, including the data symbols
    available in the lowering scope, is stored inside a `SubgraphContext` object
    that can be accessed by the visitor methods.

    This class is responsible for translation of `ir.Program`, that is the top level representation
    of a GT4Py program as a sequence of `ir.Stmt` (aka statement) expressions.
    Each statement is translated to a taskgraph inside a separate state. Statement states are chained
    one after the other: concurrency between states should be extracted by means of SDFG analysis.
    The translator will extend the SDFG while preserving the property of single exit state:
    branching is allowed within the context of one statement, but in that case the statement should
    terminate with a join state; the join state will represent the head state for next statement,
    from where to continue building the SDFG.
    """

    offset_provider_type: gtx_common.OffsetProviderType
    column_axis: Optional[gtx_common.Dimension]
    uids: gtx_utils.IDGeneratorPool = dataclasses.field(
        init=False, repr=False, default_factory=lambda: gtx_utils.IDGeneratorPool()
    )

    def get_offset_provider_type(self, offset: str) -> gtx_common.OffsetProviderTypeElem:
        return gtx_common.get_offset_type(self.offset_provider_type, offset)

    def make_field(
        self,
        data_node: dace_nodes.AccessNode,
        data_type: ts.FieldType,
    ) -> gtir_to_sdfg_types.FieldopData:
        local_dims = [dim for dim in data_type.dims if dim.kind == gtx_common.DimensionKind.LOCAL]
        if len(local_dims) == 0:
            # do nothing: the field domain consists of all global dimensions
            field_type = data_type
        elif len(local_dims) == 1:
            local_dim = local_dims[0]
            # the local dimension is converted into `ListType` data element
            if not isinstance(data_type.dtype, ts.ScalarType):
                raise ValueError(f"Invalid field type {data_type}.")
            if not gtx_common.has_offset(self.offset_provider_type, local_dim.value):
                raise ValueError(
                    f"The provided local dimension {local_dim} does not match any offset provider type."
                )
            local_type = ts.ListType(element_type=data_type.dtype, offset_type=local_dim)
            field_type = ts.FieldType(
                dims=[dim for dim in data_type.dims if dim != local_dim], dtype=local_type
            )
        else:
            raise NotImplementedError(
                "Fields with more than one local dimension are not supported."
            )
        field_origin = tuple(
            gtx_dace_args.range_start_symbol(data_node.data, dim) for dim in field_type.dims
        )
        return gtir_to_sdfg_types.FieldopData(data_node, field_type, field_origin)

    def is_column_axis(self, dim: gtx_common.Dimension) -> bool:
        assert self.column_axis
        return dim == self.column_axis

    def setup_nested_context(
        self,
        expr: gtir.Lambda,
        sdfg_name: str,
        parent_ctx: SubgraphContext,
        params: Iterable[gtir.Sym],
        symbolic_inputs: set[str],
        capture_scope_symbols: bool,
    ) -> SubgraphContext:
        assert symbolic_inputs.issubset(str(p.id) for p in params) and all(
            isinstance(p.type, ts.ScalarType) for p in params if str(p.id) in symbolic_inputs
        )

        # If `capture_scope_symbols` is True, besides the values mapped to the parameters
        # of the lambda expression (`lambda_params`), we also pass to the nested SDFG
        # all GTIR-symbols in scope, which means the symbols defined in current context.
        # Note that we first collect the symbols captured from the parent scope,
        # then we add or override symbols defined as lambda parameters.
        if capture_scope_symbols:
            lambda_symbols = {
                sym: parent_ctx.scope_symbols[sym]
                for sym in symbol_ref_utils.collect_symbol_refs(
                    expr, parent_ctx.scope_symbols.keys()
                )
            }
        else:
            lambda_symbols = {}

        # We flatten all GTIR-symbols in current scope that are captured by the
        # lambda expression, and filter the scalar symbols that are represented
        # as dace symbols in parent SDFG. These symbols have to be mapped to dace
        # symbols in the nested SDFG.
        mapped_symbols = set()
        parent_sdfg_symbols = parent_ctx.sdfg.symbols
        for name, arg_type in lambda_symbols.items():
            if isinstance(arg_type, ts.TupleType):
                mapped_symbols |= {
                    sym_id
                    for gtsym in gtir_to_sdfg_utils.flatten_tuple_fields(name, arg_type)
                    if isinstance(gtsym.type, ts.ScalarType)
                    and (sym_id := str(gtsym.id)) in parent_sdfg_symbols
                }
            elif isinstance(arg_type, ts.ScalarType) and name in parent_sdfg_symbols:
                mapped_symbols.add(name)

        # We add to 'lambda_symbols' the symbols passed as lambda parameters.
        # Note that we might eventually override symbols that already exist in
        # current scope, with the new values passed as lambda parameters.
        assert all(isinstance(p.type, ts.DataType) for p in params)
        lambda_symbols |= {
            str(p.id): p.type  # type: ignore[misc]
            for p in params
        }

        # Sorting the parameter list in alphabetical order to improve determinism.
        input_params = [
            gtir.Sym(id=name, type=lambda_symbols[name]) for name in sorted(lambda_symbols.keys())
        ]

        nsdfg = dace.SDFG(name=self.unique_nsdfg_name(sdfg_name))
        nsdfg.debuginfo = gtir_to_sdfg_utils.debug_info(expr, default=parent_ctx.sdfg.debuginfo)

        # All GTIR-symbols accessed in domain expressions by the lambda need to be
        # represented as dace symbols.
        domain_symrefs = (
            eve.walk_values(expr)
            .filter(lambda node: cpm.is_call_to(node, ("cartesian_domain", "unstructured_domain")))
            .map(
                lambda domain: eve.walk_values(domain)
                .if_isinstance(gtir.SymRef)
                .filter(lambda sym: str(sym.id) in lambda_symbols)
                .to_set()
            )
            .reduce(lambda x, y: x | y, init=set())
        )
        domain_symbols = {str(p.id) for p in _flatten_tuple_symbols(domain_symrefs)}

        # We allocate the lambda parameters as transients (see `use_transient_storage=True`).
        # When they are accessed, the corresponding data descriptor (scalar or array)
        # will be turned into global by setting `transient=False`. In this way, we remove
        # all unused (and possibly undefined) input arguments.
        self._add_sdfg_params(
            nsdfg,
            node_params=input_params,
            symbolic_params=(domain_symbols | mapped_symbols | symbolic_inputs),
            use_transient_storage=True,
        )

        nsdfg_entry_state = nsdfg.add_state(f"{sdfg_name}_entry", is_start_block=True)
        return SubgraphContext(nsdfg, nsdfg_entry_state, lambda_symbols)

    def add_nested_sdfg(
        self,
        node: gtir.Lambda,
        inner_ctx: SubgraphContext,
        outer_ctx: SubgraphContext,
        symbolic_args: Mapping[str, gtir_to_sdfg_types.SymbolicData],
        data_args: Mapping[str, gtir_to_sdfg_types.FieldopData | None],
        inner_result: gtir_to_sdfg_types.FieldopResult,
        capture_outer_data: bool,
    ) -> tuple[dace_nodes.NestedSDFG, Mapping[str, dace.Memlet]]:
        assert data_args.keys().isdisjoint(symbolic_args.keys())

        # Collect the names of all output data, by flattening any tuple structure.
        lambda_output_data = (
            gtx_utils.flatten_nested_tuple(inner_result)
            if isinstance(inner_result, tuple)
            else [inner_result]
        )
        # The output connectors only need to be setup for the actual result of the
        # internal dataflow that writes to some sink data nodes of the nested SDFG.
        lambda_outputs = {
            dataname
            for output in lambda_output_data
            if output is not None and (dataname := output.dc_node.data) not in data_args
        }

        # Map free symbols to parent SDFG
        nsdfg_symbols_mapping = {}
        for dc_symbol in inner_ctx.sdfg.free_symbols:
            if dc_symbol in data_args:
                assert (arg := data_args[dc_symbol]) is not None and isinstance(
                    arg.gt_type, ts.ScalarType
                )
                raise NotImplementedError(
                    "Unexpected mapping of scalar node to symbol on nested SDFG."
                )
            elif dc_symbol in symbolic_args:
                nsdfg_symbols_mapping[dc_symbol] = symbolic_args[dc_symbol].value
            else:
                nsdfg_symbols_mapping[dc_symbol] = dc_symbol
        for gt_symbol, arg in data_args.items():
            if arg is not None:
                nsdfg_symbols_mapping |= arg.get_symbol_mapping(gt_symbol, outer_ctx.sdfg)

        connectivity_arrays = {
            gtx_dace_args.connectivity_identifier(offset)
            for offset in gtx_dace_args.filter_connectivity_types(self.offset_provider_type)
        }

        inner_ctx_globals = [
            dataname
            for dataname, datadesc in inner_ctx.sdfg.arrays.items()
            if not datadesc.transient
        ]

        input_memlets = {}
        for dataname in inner_ctx_globals:
            if dataname in data_args:
                # Uninitialized arguments should not be used inside the nested SDFG.
                if (arg_node := data_args[dataname]) is None:
                    inner_ctx.sdfg.remove_data(dataname, validate=gtx_config.DEBUG)
                else:
                    input_memlets[dataname] = outer_ctx.sdfg.make_array_memlet(
                        arg_node.dc_node.data
                    )
            else:
                # Always capture connectivity arrays from parent scope.
                # For other GTIR-symbols (scalars, arrays), check if it is allowed.
                assert dataname in outer_ctx.sdfg.arrays
                assert dataname in connectivity_arrays or capture_outer_data
                # We check whether this global data can be removed. Besides reducing
                # the number of input connectors, this check is necessary for tuple
                # arguments, for which domain inference has detected that one or more
                # of the nested fields is not used. In such cases, the corresponding
                # argument in the top-level lambda is expected to be None (see how
                # this case is handled in the if-branch above) and it is not possible
                # to setup an input edge.
                # Note that we call `remove_data()` with `validate=True` to ensure
                # that the data is not used by any access node.
                try:
                    inner_ctx.sdfg.remove_data(dataname, validate=True)
                except ValueError:
                    # It is accessed in the lambda SDFG, so we need to setup an input edge.
                    outer_ctx.sdfg.arrays[dataname].transient = False
                    input_memlets[dataname] = outer_ctx.sdfg.make_array_memlet(dataname)

        nsdfg_node = outer_ctx.state.add_nested_sdfg(
            inner_ctx.sdfg,
            inputs=input_memlets.keys(),
            outputs=lambda_outputs,
            symbol_mapping=nsdfg_symbols_mapping,
            debuginfo=gtir_to_sdfg_utils.debug_info(node, default=outer_ctx.sdfg.debuginfo),
        )

        return nsdfg_node, input_memlets

    def unique_nsdfg_name(self, prefix: str) -> str:
        return next(self.uids[prefix])

    def unique_map_name(self, name: str) -> str:
        return f"{next(self.uids['map'])}_{name}"

    def unique_tasklet_name(self, name: str) -> str:
        return f"{next(self.uids['tlet'])}_{name}"

    def unique_temp_name(self) -> str:
        return f"{next(self.uids['gtir_tmp'])}"

    def _make_array_shape_and_strides(
        self, name: str, dims: Sequence[gtx_common.Dimension]
    ) -> tuple[list[dace.symbolic.SymbolicType], list[dace.symbolic.SymbolicType]]:
        """
        Parse field dimensions and allocate symbols for array shape and strides.

        For local dimensions, the size is known at compile-time and therefore
        the corresponding array shape dimension is set to an integer literal value.

        This method is only called for non-transient arrays, which require symbolic
        memory layout. The memory layout of transient arrays, used for temporary
        fields, is left to the DaCe default (row major, not necessarily the optimal
        one) and might be changed during optimization.

        Returns:
            Two lists of symbols, one for the shape and the other for the strides of the array.
        """
        neighbor_table_types = gtx_dace_args.filter_connectivity_types(self.offset_provider_type)
        shape = []
        for dim in dims:
            if dim.kind == gtx_common.DimensionKind.LOCAL:
                # for local dimension, the size is taken from the associated connectivity type
                shape.append(neighbor_table_types[dim.value].max_neighbors)
            elif gtx_dace_args.is_connectivity_identifier(name, self.offset_provider_type):
                # we use symbolic size for the global dimension of a connectivity
                shape.append(gtx_dace_args.field_size_symbol(name, dim, neighbor_table_types))
            else:
                # the size of global dimensions for a regular field is the symbolic
                # expression of domain range 'stop - start'
                shape.append(
                    dace.symbolic.pystr_to_symbolic(
                        "{} - {}".format(
                            gtx_dace_args.range_stop_symbol(name, dim),
                            gtx_dace_args.range_start_symbol(name, dim),
                        )
                    )
                )
        strides = [
            gtx_dace_args.field_stride_symbol(name, dim, neighbor_table_types) for dim in dims
        ]
        return shape, strides

    def _add_storage(
        self,
        sdfg: dace.SDFG,
        symbolic_params: set[str] | None,
        name: str,
        gt_type: ts.DataType,
        transient: bool,
    ) -> list[tuple[str, ts.DataType]]:
        """
        Add storage in the SDFG for a given GT4Py data symbol.

        GT4Py fields are allocated as DaCe arrays. GT4Py scalars are represented
        as scalar data containers, unless they are listed in `symbolic_arguments`,
        in which case they are represented as SDFG symbols. This is the case of
        start/stop symbols for field range or other scalar values accessed in
        domain symbolic expressions.

        The fields used as temporary arrays, when `transient = True`, are allocated
        and exist only within the SDFG; when `transient = False`, the fields have
        to be allocated outside and have to be passed as arguments to the SDFG call.

        Args:
            sdfg: The SDFG where storage needs to be allocated.
            symbolic_params: GT4Py scalars that must be represented as SDFG symbols.
                If `None`, all scalar parameters should be represented as dace symbols.
            name: Symbol Name to be allocated.
            gt_type: GT4Py symbol type.
            transient: True when the data symbol has to be allocated as internal storage.

        Returns:
            List of tuples '(data_name, gt_type)' where 'data_name' is the name of
            the data container used as storage in the SDFG and 'gt_type' is the
            corresponding GT4Py type. In case the storage has to be allocated for
            a tuple symbol the list contains a flattened version of the tuple,
            otherwise the list will contain a single entry.
        """
        if isinstance(gt_type, ts.TupleType):
            tuple_fields = []
            for sym in gtir_to_sdfg_utils.flatten_tuple_fields(name, gt_type):
                assert isinstance(sym.type, ts.DataType)
                tuple_fields.extend(
                    self._add_storage(
                        sdfg=sdfg,
                        symbolic_params=symbolic_params,
                        name=str(sym.id),
                        gt_type=sym.type,
                        transient=transient,
                    )
                )
            return tuple_fields

        elif isinstance(gt_type, ts.FieldType):
            if len(gt_type.dims) == 0:
                # represent zero-dimensional fields as scalar arguments
                return self._add_storage(
                    sdfg=sdfg,
                    symbolic_params=set(),  # force use of scalar data container
                    name=name,
                    gt_type=gt_type.dtype,
                    transient=transient,
                )
            if isinstance(gt_type.dtype, ts.ScalarType):
                dc_dtype = gtx_dace_args.as_dace_type(gt_type.dtype)
                all_dims = gt_type.dims
            else:  # for 'ts.ListType' use 'offset_type' as local dimension
                assert gt_type.dtype.offset_type is not None
                assert gt_type.dtype.offset_type.kind == gtx_common.DimensionKind.LOCAL
                assert isinstance(gt_type.dtype.element_type, ts.ScalarType)
                dc_dtype = gtx_dace_args.as_dace_type(gt_type.dtype.element_type)
                all_dims = gtx_common.order_dimensions([*gt_type.dims, gt_type.dtype.offset_type])

            # Use symbolic shape, which allows to invoke the program with fields of different size;
            # and symbolic strides, which enables decoupling the memory layout from generated code.
            sym_shape, sym_strides = self._make_array_shape_and_strides(name, all_dims)
            sdfg.add_array(name, sym_shape, dc_dtype, strides=sym_strides, transient=transient)
            return [(name, gt_type)]

        elif isinstance(gt_type, ts.ScalarType):
            dc_dtype = gtx_dace_args.as_dace_type(gt_type)
            if symbolic_params is None or name in symbolic_params:
                sdfg.add_symbol(name, dc_dtype)
            else:
                sdfg.add_scalar(name, dc_dtype, transient=transient)

            return [(name, gt_type)]

        raise RuntimeError(f"Data type '{type(gt_type)}' not supported.")

    def _visit_expression(
        self,
        node: gtir.Expr,
        ctx: SubgraphContext,
        use_temp: bool = True,
    ) -> gtir_to_sdfg_types.FieldopResult:
        """
        Specialized visit method for fieldview expressions.

        This method represents the entry point to visit `ir.Stmt` expressions.
        As such, it must preserve the property of single exit state in the SDFG.

        Args:
            node: The GTIR expression to be lowered.
            sdfg: The SDFG which is being constructed.
            head_state: The state inside the given SDFG where the GTIR expression
                should be lowered.
            use_temp: If True, this method ensures that the result data is written
                to temporary storage.

        Returns:
            The SDFG array nodes containing the result of the fieldview expression.
            The nodes are organized in tree form, in case of tuples.
        """

        result = self.visit(node, ctx=ctx)

        # sanity check: each statement should preserve the property of single exit state (aka head state),
        # i.e. eventually only introduce internal branches, and keep the same head state
        sink_states = ctx.sdfg.sink_nodes()
        assert len(sink_states) == 1
        assert sink_states[0] == ctx.state

        if use_temp:  # copy the full shape of global data to temporary storage
            return gtx_utils.tree_map(
                lambda x: x
                if x is None or x.dc_node.desc(ctx.sdfg).transient
                else ctx.copy_data(self, x, domain=None)
            )(result)
        else:
            return result

    def _add_sdfg_params(
        self,
        sdfg: dace.SDFG,
        node_params: Sequence[gtir.Sym],
        symbolic_params: set[str] | None,
        use_transient_storage: bool,
    ) -> list[str]:
        """
        Helper function to add storage for node parameters and connectivity tables.

        If `use_transient_storage=True`, all scalars and arrays are allocated as
        transient data. When this data is accessed by the `SymRef` visitor or mapped
        to input data inside a nested SDFG, during lowering of let-lambdas, it is
        turned into global. This allows, for nested SDFGs, to prune input connectors
        for data that is not used. The remaining unused transients in the SDFG are
        removed by the dace simplify pass. For connectivity arrays, we always use
        this approach, indipendently from the `use_transient_storage` argument.

        For details about storage allocation of each data type, see `_add_storage()`.
        """

        # add non-transient arrays and/or SDFG symbols for the program arguments
        sdfg_args = []
        for param in node_params:
            gt_symbol_name = str(param.id)
            assert isinstance(param.type, (ts.DataType))
            sdfg_args += self._add_storage(
                sdfg=sdfg,
                symbolic_params=symbolic_params,
                name=gt_symbol_name,
                gt_type=param.type,
                transient=use_transient_storage,
            )

        # add SDFG storage for connectivity tables
        for offset, connectivity_type in gtx_dace_args.filter_connectivity_types(
            self.offset_provider_type
        ).items():
            gt_type = ts.FieldType(
                dims=[connectivity_type.source_dim, connectivity_type.neighbor_dim],
                dtype=tt.from_dtype(connectivity_type.dtype),
            )
            # We store all connectivity tables as transient arrays here; later, while building
            # the field operator expressions, we change to non-transient (i.e. allocated externally)
            # the tables that are actually used. This way, we avoid adding SDFG arguments for
            # the connectivity tables that are not used. The remaining unused transient arrays
            # are removed by the dace simplify pass.
            self._add_storage(
                sdfg=sdfg,
                symbolic_params=symbolic_params,
                name=gtx_dace_args.connectivity_identifier(offset),
                gt_type=gt_type,
                transient=True,
            )

        # the list of all sdfg arguments (aka non-transient arrays) which include tuple-element fields
        return [arg_name for arg_name, _ in sdfg_args]

    def visit_Program(self, node: gtir.Program) -> dace.SDFG:
        """Translates `ir.Program` to `dace.SDFG`.

        First, it will allocate field and scalar storage for global data. The storage
        represents global data, available everywhere in the SDFG, either containing
        external data (aka non-transient data) or temporary data (aka transient data).
        The temporary data is global, therefore available everywhere in the SDFG
        but not outside. Then, all statements are translated, one after the other.
        """
        sdfg = dace.SDFG(str(node.id))
        sdfg.debuginfo = gtir_to_sdfg_utils.debug_info(node)

        # start block of the stateful graph
        head_state = sdfg.add_state("program_entry", is_start_block=True)

        # By passing `symbolic_arguments=None` all scalars are represented as dace symbols.
        #   We do this to allow lowering of scalar expressions in let-statements,
        #   that only depend on scalar parameters, as dace symbolic expressions
        #   mapped to symbols on a nested SDFG.
        sdfg_arg_names = self._add_sdfg_params(
            sdfg, node.params, symbolic_params=None, use_transient_storage=False
        )

        # visit one statement at a time and expand the SDFG from the current head state
        scope_symbols = {str(p.id): p.type for p in node.params if isinstance(p.type, ts.DataType)}
        for i, stmt in enumerate(node.body):
            # include `debuginfo` only for `ir.Program` and `ir.Stmt` nodes: finer granularity would be too messy
            head_state = sdfg.add_state_after(head_state, f"stmt_{i}")
            head_state._debuginfo = gtir_to_sdfg_utils.debug_info(stmt, default=sdfg.debuginfo)
            head_state = self.visit(stmt, ctx=SubgraphContext(sdfg, head_state, scope_symbols))

        # remove unused connectivity tables (by design, arrays are marked as non-transient when they are used)
        for nsdfg in sdfg.all_sdfgs_recursive():
            unused_connectivities = [
                data
                for data, datadesc in nsdfg.arrays.items()
                if gtx_dace_args.is_connectivity_identifier(data, self.offset_provider_type)
                and datadesc.transient
            ]
            for data in unused_connectivities:
                assert isinstance(nsdfg.arrays[data], dace.data.Array)
                nsdfg.arrays.pop(data)

        # Create the call signature for the SDFG.
        #  Only the arguments required by the GT4Py program, i.e. `node.params`, are added
        #  as positional arguments. The implicit arguments, such as the offset providers or
        #  the arguments created by the translation process, must be passed as keyword arguments.
        sdfg.arg_names = sdfg_arg_names

        return sdfg

    def visit_SetAt(self, stmt: gtir.SetAt, ctx: SubgraphContext) -> dace.SDFGState:
        """Visits a `SetAt` statement expression and writes the local result to some external storage.

        Each statement expression results in some sort of dataflow gragh writing to temporary
        storage, inside the current state. The translation of `SetAt` ensures that the result
        is written back to the target external storage, in a new state after the current one,
        which will become the new head state.

        Returns:
          The new SDFG head state, where the writes to the AccessNodes of the field operator results are located.
        """

        def _visit_target(
            source: gtir_to_sdfg_types.FieldopData,
            target: gtir_to_sdfg_types.FieldopData,
            target_domain: domain_utils.SymbolicDomain,
            target_state: dace.SDFGState,
            target_is_also_used_as_input: bool,
        ) -> None:
            assert source.dc_node.desc(ctx.sdfg).transient
            assert source.gt_type == target.gt_type
            field_domain = gtir_domain.get_field_domain(target_domain)
            source_subset = _make_access_index_for_field(field_domain, source)
            target_subset = _make_access_index_for_field(field_domain, target)

            if target_is_also_used_as_input:
                # write the field operator result in the separate target state,
                # in order to ensure the correct dataflow for write after read
                target_state.add_nedge(
                    target_state.add_access(source.dc_node.data),
                    target.dc_node,
                    dace.Memlet(
                        data=source.dc_node.data, subset=source_subset, other_subset=target_subset
                    ),
                )
            else:
                # do not use the target state, write the result inside the current context
                ctx.state.add_nedge(
                    source.dc_node,
                    ctx.state.add_access(target.dc_node.data),
                    dace.Memlet(
                        data=source.dc_node.data, subset=source_subset, other_subset=target_subset
                    ),
                )
                target_state.remove_node(target.dc_node)

        # Visit the domain expression.
        domain = gtir_domain.extract_target_domain(stmt.domain)

        # Visit the field operator expression.
        source_tree = self._visit_expression(stmt.expr, ctx)
        ctx_input_data = ctx.get_input_data()

        # In order to support inout argument, write the result in separate next state
        # this is needed to avoid indeterministic behavior for expressions like: X, Y = X + 1, X
        target_state = ctx.sdfg.add_state_after(ctx.state, f"post_{ctx.state.label}")
        # The target expression could be a `SymRef` to an output field or a `make_tuple`
        # expression in case the statement returns more than one field.
        target_tree = self._visit_expression(
            stmt.target,
            ctx=SubgraphContext(ctx.sdfg, target_state, ctx.scope_symbols),
            use_temp=False,
        )
        gtx_utils.tree_map(
            lambda source, target, target_domain: _visit_target(
                source,
                target,
                target_domain,
                target_state,
                target_is_also_used_as_input=(target.dc_node.data in ctx_input_data),
            )
        )(source_tree, target_tree, domain)

        if target_state.is_empty():
            ctx.sdfg.remove_node(target_state)
            return ctx.state
        else:
            return target_state

    def visit_FunCall(
        self,
        node: gtir.FunCall,
        ctx: SubgraphContext,
    ) -> gtir_to_sdfg_types.FieldopResult:
        # use specialized dataflow builder classes for each builtin function
        if cpm.is_call_to(node, "concat_where"):
            return gtir_to_sdfg_concat_where.translate_concat_where(node, ctx, self)
        elif cpm.is_call_to(node, "if_"):
            return gtir_to_sdfg_primitives.translate_if(node, ctx, self)
        elif cpm.is_call_to(node, "index"):
            return gtir_to_sdfg_primitives.translate_index(node, ctx, self)
        elif cpm.is_call_to(node, "make_tuple"):
            return gtir_to_sdfg_primitives.translate_make_tuple(node, ctx, self)
        elif cpm.is_call_to(node, "tuple_get"):
            return gtir_to_sdfg_primitives.translate_tuple_get(node, ctx, self)
        elif cpm.is_applied_as_fieldop(node):
            return gtir_to_sdfg_primitives.translate_as_fieldop(node, ctx, self)
        elif isinstance(node.fun, gtir.Lambda):
            # Special handling of scalar arguments of a let-lambda that can be lowered
            # as symbolic expressions: when all the GTIR-symbols the scalar expression
            # depends on are dace symbols, the argument can be passed to the nested
            # SDFG by means of symbol mapping.
            symbolic_args = {}
            for p, lambda_arg in zip(node.fun.params, node.args, strict=True):
                if not isinstance(lambda_arg.type, ts.ScalarType):
                    continue
                # Convert the scalar argument to a dace symbolic expression if all
                # of its dependencies are symbols to.
                try:
                    symbolic_expr = gtir_to_sdfg_utils.get_symbolic(lambda_arg)
                except TypeError:
                    # sympy parsing failed, it can happen with 'cast_' expressions
                    if not any(
                        eve.walk_values(lambda_arg).map(lambda node: cpm.is_call_to(node, "cast_"))
                    ):
                        raise
                    continue
                if all(str(s) in ctx.sdfg.symbols for s in symbolic_expr.free_symbols):
                    symbolic_args[str(p.id)] = symbolic_expr
            # All other lambda arguments are lowered to some dataflow that produces a data node.
            args = {
                param: (
                    gtir_to_sdfg_types.SymbolicData(param.type, symbolic_args[gt_symbol_name])  # type: ignore[arg-type]
                    if (gt_symbol_name := str(param.id)) in symbolic_args
                    else self.visit(arg, ctx=ctx)
                )
                for param, arg in zip(node.fun.params, node.args, strict=True)
            }
            return self.visit(node.fun, ctx=ctx, args=args)
        elif isinstance(node.type, ts.ScalarType):
            return gtir_to_sdfg_primitives.translate_scalar_expr(node, ctx, self)
        else:
            raise NotImplementedError(f"Unexpected 'FunCall' expression ({node}).")

    def visit_Lambda(
        self,
        node: gtir.Lambda,
        ctx: SubgraphContext,
        args: Mapping[gtir.Sym, gtir_to_sdfg_types.FieldopResult | gtir_to_sdfg_types.SymbolicData],
    ) -> gtir_to_sdfg_types.FieldopResult:
        """
        Translates a `Lambda` node to a nested SDFG in the current state.

        The reason for creating a nested SDFG is to define local symbols (the lambda
        paremeters) that map to parent fields, either program arguments, temporaries
        or symbolic expressions.

        The arguments passed to the lambda expression are divided in two groups:
        * arguments passed as access nodes, i.e. `as_fieldop` result, field and scalar
          `gtir.SymRef`, or nested let-lambdas;
        * arguments (scalars only) that can be evaluated as symbolic expressions,
          which can be mapped to symbols on the nested SDFG.


        If the lambda has a parameter whose name is already present in `GTIRToSDFG.global_symbols`,
        i.e. a lambda parameter with the same name as a symbol in scope, the parameter will shadow
        the previous symbol during traversal of the lambda expression.
        """

        lambda_arg_nodes, symbolic_args = flatten_tuple_args(args.items())

        # lower let-statement lambda node as a nested SDFG
        lambda_ctx = self.setup_nested_context(
            expr=node,
            sdfg_name="lambda",
            parent_ctx=ctx,
            params=args.keys(),
            symbolic_inputs=set(symbolic_args.keys()),
            capture_scope_symbols=True,
        )

        lambda_result = self.visit(node.expr, ctx=lambda_ctx)

        # A let-lambda is allowed to capture GTIR-symbols from the outer scope,
        # therefore we call `add_nested_sdfg()` with `capture_outer_data=True`.
        nsdfg_node, input_memlets = self.add_nested_sdfg(
            node=node,
            inner_ctx=lambda_ctx,
            outer_ctx=ctx,
            symbolic_args=symbolic_args,
            data_args=lambda_arg_nodes,
            inner_result=lambda_result,
            capture_outer_data=True,
        )

        # In this loop we call `pop()`, whenever an argument is connected to an input
        # connector on the nested SDFG, so the corresponding node is removed from the dictionary.
        for input_connector, memlet in input_memlets.items():
            if input_connector in lambda_arg_nodes:
                arg_node = lambda_arg_nodes.pop(input_connector)
                assert arg_node is not None
                src_node = arg_node.dc_node
            else:
                src_node = ctx.state.add_access(memlet.data)

            ctx.state.add_edge(src_node, None, nsdfg_node, input_connector, memlet)

        # We can now safely remove all remaining arguments, because unused.
        # It can happen that a let-lambda argument is not used, in which case it
        # should be `None` at this stage, but the symbol passed as argument is captured
        # by the let-lambda expression and thus can be accessed directly inside the
        # inner scope. The domain in the annex of this symbol is not `NEVER`, and
        # an access node is created, so we find an isolated access node here.
        if unused_access_nodes := [
            arg_node.dc_node
            for arg_node in lambda_arg_nodes.values()
            if not (arg_node is None or arg_node.dc_node.desc(ctx.sdfg).transient)
        ]:
            assert all(ctx.state.degree(access_node) == 0 for access_node in unused_access_nodes)
            ctx.state.remove_nodes_from(unused_access_nodes)

        def construct_output_for_nested_sdfg(
            inner_data: gtir_to_sdfg_types.FieldopData,
        ) -> gtir_to_sdfg_types.FieldopData:
            """
            This function makes a data container that lives inside a nested SDFG, denoted by `inner_data`,
            available in the parent SDFG.
            In order to achieve this, the data container inside the nested SDFG is marked as non-transient
            (in other words, externally allocated - a requirement of the SDFG IR) and a new data container
            is created within the parent SDFG, with the same properties (shape, stride, etc.) of `inner_data`
            but appropriatly remapped using the symbol mapping table.
            For lambda arguments that are simply returned by the lambda, the `inner_data` was already mapped
            to a parent SDFG data container, therefore it can be directly accessed in the parent SDFG.
            The same happens to symbols available in the lambda context but not explicitly passed as lambda
            arguments, that are simply returned by the lambda: it can be directly accessed in the parent SDFG.
            """
            if not inner_data.dc_node.desc(lambda_ctx.sdfg).transient:
                # This is inout data: some global associated to an input connector
                # is also a sink node of the lambda dataflow. This can happen, for
                # example, when the lambda constructs a tuple of some input fields.
                # We copy this data to a new node, which we use as output.
                nsdfg_node.remove_out_connector(inner_data.dc_node.data)
                inner_data = lambda_ctx.copy_data(self, inner_data, domain=None)
                nsdfg_node.add_out_connector(inner_data.dc_node.data)
            elif lambda_ctx.state.degree(inner_data.dc_node) == 0:
                # Isolated access node will make validation fail.
                # Isolated access nodes can be found in the join-state of an if-expression.
                lambda_ctx.state.remove_node(inner_data.dc_node)
            # Transient data nodes only exist within the nested SDFG. In order to return some result data,
            # the corresponding data container inside the nested SDFG has to be changed to non-transient,
            # that is externally allocated, as required by the SDFG IR. An output edge will write the result
            # from the nested-SDFG to a new intermediate data container allocated in the parent SDFG.
            outer_data = ctx.map_nsdfg_field(
                sdfg_builder=self,
                nsdfg_field=inner_data,
                nsdfg=lambda_ctx.sdfg,
                symbol_mapping=nsdfg_node.symbol_mapping,
            )
            ctx.state.add_edge(
                nsdfg_node,
                inner_data.dc_node.data,
                outer_data.dc_node,
                None,
                ctx.sdfg.make_array_memlet(outer_data.dc_node.data),
            )

            return outer_data

        return gtx_utils.tree_map(construct_output_for_nested_sdfg)(lambda_result)

    def visit_Literal(
        self,
        node: gtir.Literal,
        ctx: SubgraphContext,
    ) -> gtir_to_sdfg_types.FieldopResult:
        return gtir_to_sdfg_primitives.translate_literal(node, ctx, self)

    def visit_SymRef(
        self,
        node: gtir.SymRef,
        ctx: SubgraphContext,
    ) -> gtir_to_sdfg_types.FieldopResult:
        return gtir_to_sdfg_primitives.translate_symbol_ref(node, ctx, self)


def build_sdfg_from_gtir(
    ir: gtir.Program,
    offset_provider_type: gtx_common.OffsetProviderType,
    column_axis: Optional[gtx_common.Dimension] = None,
) -> dace.SDFG:
    """
    Receives a GTIR program and lowers it to a DaCe SDFG.

    The lowering to SDFG requires that the program node is type-annotated, therefore
    this function runs type ineference as first step.

    Args:
        ir: The GTIR program node to be lowered to SDFG
        offset_provider_type: The definitions of offset providers used by the program node
        column_axis: Vertical dimension used for column scan expressions.

    Returns:
        An SDFG in the DaCe canonical form (simplified)
    """

    if ir.function_definitions:
        raise NotImplementedError("Functions expected to be inlined as lambda calls.")
    if ir.declarations:
        raise NotImplementedError("Temporaries not supported yet by GTIR DaCe backend.")

    ir = gtir_type_inference.infer(ir, offset_provider_type=offset_provider_type)
    ir = ir_prune_casts.PruneCasts().visit(ir)

    # DaCe requires C-compatible strings for the names of data containers,
    # such as arrays and scalars. GT4Py uses a unicode symbols ('ᐞ') as name
    # separator in the SSA pass, which generates invalid symbols for DaCe.
    # Here we find new names for invalid symbols present in the IR.
    ir = gtir_to_sdfg_utils.replace_invalid_symbols(ir)

    sdfg_genenerator = GTIRToSDFG(offset_provider_type, column_axis)
    sdfg = sdfg_genenerator.visit(ir)
    assert isinstance(sdfg, dace.SDFG)

    sdfg.validate()
    return sdfg



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from gt4py.cartesian import backend
from gt4py.cartesian.backend import python_common as py_common
from gt4py.cartesian.gtc import numpy, passes
from gt4py.cartesian.gtc.gtir_to_oir import GTIRToOIR
from gt4py.cartesian.gtc.numpy import npir
from gt4py.cartesian.gtc.passes import oir_optimizations as oir_opt
from gt4py.eve import codegen
from gt4py.storage import layout


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_object import StencilObject


@backend.register
class NumpyBackend(backend.BaseBackend):
    """NumPy backend using gtc."""

    name = "numpy"
    options: ClassVar[dict[str, Any]] = {
        "oir_pipeline": {"versioning": True, "type": passes.OirPipeline},
        # TODO: Implement this option in source code
        "ignore_np_errstate": {"versioning": True, "type": bool},
    }
    storage_info = layout.NaiveCPULayout
    languages: ClassVar[dict] = {"computation": "python", "bindings": ["python"]}
    MODULE_GENERATOR_CLASS = py_common.PythonModuleGenerator

    def generate_computation(self) -> dict[str, str | dict]:
        ignore_np_errstate = self.builder.options.backend_opts.get("ignore_np_errstate", True)
        source = numpy.NpirCodegen.apply(self.npir, ignore_np_errstate=ignore_np_errstate)

        if self.builder.options.format_source:
            source = codegen.format_source("python", source)

        caching = self.builder.caching
        computation_name = f"{caching.module_prefix}computation{caching.module_postfix}.py"
        return {computation_name: source}

    def generate(self) -> type[StencilObject]:
        self.check_options(self.builder.options)
        src_dir = self.builder.module_path.parent

        if not self.builder.options._impl_opts.get("disable-code-generation", False):
            py_common.recursive_write(src_dir, self.generate_computation())

        return self.make_module()

    def _make_npir(self) -> npir.Computation:
        base_oir = GTIRToOIR().visit(self.builder.gtir)
        oir_pipeline = self.builder.options.backend_opts.get(
            "oir_pipeline",
            passes.DefaultPipeline(
                skip=[
                    oir_opt.IJCacheDetection,
                    oir_opt.KCacheDetection,
                    oir_opt.PruneKCacheFills,
                    oir_opt.PruneKCacheFlushes,
                ]
            ),
        )
        oir_node = oir_pipeline.run(base_oir)
        base_npir = numpy.OirToNpir().visit(oir_node)
        return numpy.ScalarsToTemporaries().visit(base_npir)

    @property
    def npir(self) -> npir.Computation:
        key = "gtcnumpy:npir"
        if key not in self.builder.backend_data:
            self.builder.with_backend_data({key: self._make_npir()})
        return self.builder.backend_data[key]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from typing import Any, Sequence

import numpy as np

from gt4py._core import definitions as core_defs
from gt4py.next import common as gtx_common, config, utils as gtx_utils
from gt4py.next.instrumentation import metrics
from gt4py.next.otf import stages
from gt4py.next.program_processors.runners.dace import sdfg_callable
from gt4py.next.program_processors.runners.dace.workflow import (
    common as gtx_wfdcommon,
    compilation as gtx_wfdcompilation,
)


def convert_args(
    fun: gtx_wfdcompilation.CompiledDaceProgram,
    device: core_defs.DeviceType = core_defs.DeviceType.CPU,
) -> stages.ExecutableProgram:
    # Retieve metrics level from GT4Py environment variable.
    collect_time = metrics.is_level_enabled(metrics.PERFORMANCE)
    collect_time_arg = np.array([1], dtype=np.float64)
    # We use the callback function provided by the compiled program to update the SDFG arglist.
    update_sdfg_call_args = functools.partial(
        fun.update_sdfg_ctype_arglist, device, fun.sdfg_argtypes
    )

    def decorated_program(
        *args: Any,
        offset_provider: gtx_common.OffsetProvider,
        out: Any = None,
    ) -> Any:
        if out is not None:
            args = (*args, out)

        try:
            # Not the first call.
            #  We will only update the argument vector  for the normal call.
            # NOTE: If this is the first time then we will generate an exception because
            #   `fun.csdfg_args` is `None`
            # TODO(phimuell, edopao): Think about refactor the code such that the update
            #   of the argument vector is a Method of the `CompiledDaceProgram`.
            update_sdfg_call_args(args, fun.csdfg_argv, offset_provider)  # type: ignore[arg-type]  # Will error out in first call.

        except TypeError:
            # First call. Construct the initial argument vector of the `CompiledDaceProgram`.
            assert fun.csdfg_argv is None and fun.csdfg_init_argv is None
            flat_args: Sequence[Any] = gtx_utils.flatten_nested_tuple(args)
            this_call_args = sdfg_callable.get_sdfg_args(
                fun.sdfg_program.sdfg,
                offset_provider,
                *flat_args,
                filter_args=False,
            )
            this_call_args |= {
                gtx_wfdcommon.SDFG_ARG_METRIC_LEVEL: config.COLLECT_METRICS_LEVEL,
                gtx_wfdcommon.SDFG_ARG_METRIC_COMPUTE_TIME: collect_time_arg,
            }
            fun.construct_arguments(**this_call_args)

        # Perform the call to the SDFG.
        fun.fast_call()

        if collect_time:
            metrics.add_sample_to_current_source(metrics.COMPUTE_METRIC, collect_time_arg[0].item())

    return decorated_program



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dace.sdfg.analysis.schedule_tree import treenodes as tn

from gt4py.cartesian.gtc.dace.passes import utils


class PushVerticalMapDown(tn.ScheduleNodeVisitor):
    """
    Given a schedule tree with K-JI loops, push the K-loop down into the JI-loops to
    achieve a loop structure suitable for a "K-first" memory layout, or a layout
    where K is the highest value in the layout_map.

    Expected input is something like

    // vertical loop outside
    for k ...
        // horizontal loop(s) inside
        for j,i ...
            // computation here (1)

        for j,i ...
            // computation here (2)

    Given such a loop structure, this pass will push down the vertical loop into the
    horizontal loops. In the above example, the expected output is

    // horizontal loop(s) outside
    for j,i ...
        for k ...
            // computation here (1)

    for j,i ...
        for k ...
            // computation here (2)
    """

    def _push_K_loop_in_IJ(self, node: tn.MapScope | tn.ForScope):
        # take refs before moving things around
        parent = node
        grandparent = node.parent
        grandparent_children = node.parent.children
        k_loop_index = utils.list_index(grandparent_children, parent)

        for child in node.children:
            if not isinstance(child, tn.MapScope):
                raise NotImplementedError("We don't expect anything else than (IJ)-MapScopes here.")

            # New loop with MapEntry (`node`) from parent and children from `child`
            if isinstance(node, tn.MapScope):
                new_loop = tn.MapScope(node=parent.node, children=child.children)
                new_loop.parent = child
            elif isinstance(node, tn.ForScope):
                new_loop = node
                node.children = child.children
                node.parent = child
            else:
                raise ValueError(f"Unknown node of type {type(node)}")
            child.children = [new_loop]
            child.parent = grandparent
            grandparent_children.insert(k_loop_index, child)
            k_loop_index += 1

        # delete old (now unused) node
        grandparent_children.remove(node)

    def visit_MapScope(self, node: tn.MapScope):
        if node.node.map.params[0].startswith("__k"):
            self._push_K_loop_in_IJ(node)

    def visit_ForScope(self, node: tn.ForScope):
        if node.loop.loop_variable.startswith("__k"):
            self._push_K_loop_in_IJ(node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
import dataclasses
from typing import (
    Any,
    Dict,
    Final,
    Iterable,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeAlias,
    Union,
)

import dace
from dace import nodes as dace_nodes, subsets as dace_subsets

from gt4py import eve
from gt4py.eve.extended_typing import MaybeNestedInTuple, NestedTuple
from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import symbol_ref_utils
from gt4py.next.program_processors.runners.dace import sdfg_args as gtx_dace_args
from gt4py.next.program_processors.runners.dace.lowering import (
    gtir_python_codegen,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
)
from gt4py.next.type_system import (
    type_info as ti,
    type_specifications as ts,
    type_translation as tt,
)


# Magic local dimension used for list of values with length known at compile-time.
_CONST_DIM: Final = gtx_common.Dimension(value="_CONST_DIM", kind=gtx_common.DimensionKind.LOCAL)


@dataclasses.dataclass(frozen=True)
class ValueExpr:
    """
    Local storage for the values returned by dataflow computation.

    This type is used in the context in a dataflow, that is a stencil expression.
    Therefore, it contains either a scalar value (single elements in the fields) or
    a list of values in a local dimension.
    This is different from `gtir_to_sdfg_types.FieldopData` which represents
    the result of a field operator, basically the data storage outside a global map.

    Args:
        dc_node: Access node to the data container, can be either a scalar or a local list.
        gt_dtype: GT4Py data type, which includes the `offset_type` local dimension for lists.
    """

    dc_node: dace_nodes.AccessNode
    gt_dtype: ts.ListType | ts.ScalarType

    def __post_init__(self) -> None:
        if isinstance(self.gt_dtype, ts.ListType):
            assert self.gt_dtype.offset_type is not None
            assert self.gt_dtype.offset_type.kind == gtx_common.DimensionKind.LOCAL


@dataclasses.dataclass(frozen=True)
class MemletExpr:
    """
    Memlet to retrieve local data, either a scalar or a list, from a field.

    Args:
        dc_node: Access node to the array data container of the source field.
        gt_field: GT4Py field type definition, which includes the list of dimensions
          according to the canonical field layout, used to modify the memlet subset.
        subset: The memlet subset to retrieve the local data.
    """

    dc_node: dace_nodes.AccessNode
    gt_field: ts.FieldType
    subset: dace_subsets.Range

    @property
    def gt_dtype(self) -> ts.ScalarType | ts.ListType:
        return self.gt_field.dtype

    def __post_init__(self) -> None:
        if isinstance(self.gt_dtype, ts.ListType):
            assert self.gt_dtype.offset_type is not None
            assert self.gt_dtype.offset_type.kind == gtx_common.DimensionKind.LOCAL


@dataclasses.dataclass(frozen=True)
class SymbolExpr:
    """Any symbolic expression that is constant in the context of current SDFG."""

    value: dace.symbolic.SymbolicType
    dc_dtype: dace.typeclass


DataExpr: TypeAlias = ValueExpr | MemletExpr | SymbolExpr


@dataclasses.dataclass(frozen=True)
class IteratorExpr:
    """
    Iterator for field access to be consumed by `deref` or `shift` builtin functions.

    Args:
        field: Access node to the field this iterator operates on.
        gt_dtype: GT4Py data type, which includes the `offset_type` local dimension for lists.
        field_domain: Field domain represented as a sorted list of dimensions and offset values,
            used to find the position of a map index variable in the memlet subset. The offset
            value is either the start index of dimension range or the compile-time value of
            a shift expression, or a composition of both, and it must be subtracted to the index
            variable when constructing the memlet subset range.
        indices: Maps each dimension to an index value, which could be either a symbolic value
            or the result of a tasklet computation like neighbors connectivity or dynamic offset.
    """

    field: dace_nodes.AccessNode
    gt_dtype: ts.ListType | ts.ScalarType
    field_domain: list[tuple[gtx_common.Dimension, dace.symbolic.SymbolicType]]
    indices: dict[gtx_common.Dimension, DataExpr]

    def __post_init__(self) -> None:
        gtx_common.check_dims([dim for dim, _ in self.field_domain])
        if isinstance(self.gt_dtype, ts.ListType):
            assert self.gt_dtype.offset_type is not None
            assert self.gt_dtype.offset_type.kind == gtx_common.DimensionKind.LOCAL
            assert all(dim != self.gt_dtype.offset_type for dim, _ in self.field_domain)

    def get_field_type(self) -> ts.FieldType:
        return ts.FieldType(dims=[dim for dim, _ in self.field_domain], dtype=self.gt_dtype)

    def get_memlet_subset(self, sdfg: dace.SDFG) -> dace_subsets.Range:
        if len(self.field_domain) == 0:  # zero-dimensional field
            return dace_subsets.Range.from_string("0")

        if not all(isinstance(self.indices[dim], SymbolExpr) for dim, _ in self.field_domain):
            raise ValueError(f"Cannot deref iterator {self}.")

        field_desc = self.field.desc(sdfg)
        if isinstance(self.gt_dtype, ts.ListType):
            assert len(field_desc.shape) == len(self.field_domain) + 1
            local_dim: gtx_common.Dimension = self.gt_dtype.offset_type  # type: ignore[assignment]  # checked in __post_init__
            # construct the full subset according to the canonical field domain
            sorted_dims = gtx_common.order_dimensions(
                [dim for dim, _ in self.field_domain] + [local_dim]
            )
            local_dim_index = sorted_dims.index(local_dim)
            field_domain = [
                *self.field_domain[:local_dim_index],
                (local_dim, 0),
                *self.field_domain[local_dim_index:],
            ]
        else:
            assert len(field_desc.shape) == len(self.field_domain)
            field_domain = self.field_domain

        return dace_subsets.Range.from_string(
            ",".join(
                str(self.indices[dim].value - offset)  # type: ignore[union-attr]
                if dim in self.indices
                else f"0:{size}"
                for (dim, offset), size in zip(field_domain, field_desc.shape, strict=True)
            )
        )


class DataflowInputEdge(Protocol):
    """
    This protocol describes how to concretize a data edge to read data from a source node
    into the dataflow.

    It provides the `connect` method to setup an input edge from an external data source.
    The most common case is that the dataflow represents a stencil, which is instantied
    inside a map scope and whose inputs and outputs are connected to external data nodes
    by means of memlets that traverse the map entry and exit nodes.
    The dataflow can also be instatiated without a map, in which case the `map_entry`
    argument is set to `None`.
    """

    @abc.abstractmethod
    def connect(self, map_entry: Optional[dace_nodes.MapEntry]) -> None: ...


@dataclasses.dataclass(frozen=True)
class MemletInputEdge(DataflowInputEdge):
    """
    Allows to setup an input memlet through a map entry node.

    The edge source has to be a data access node, while the destination node can either
    be a tasklet, in which case the connector name is also required, or an access node.
    """

    state: dace.SDFGState
    source: dace_nodes.AccessNode
    subset: dace_subsets.Range
    dest: dace_nodes.AccessNode | dace_nodes.Tasklet
    dest_conn: Optional[str]

    def connect(self, map_entry: Optional[dace_nodes.MapEntry]) -> None:
        memlet = dace.Memlet(data=self.source.data, subset=self.subset)
        if map_entry is None:
            self.state.add_edge(self.source, None, self.dest, self.dest_conn, memlet)
        else:
            self.state.add_memlet_path(
                self.source,
                map_entry,
                self.dest,
                dst_conn=self.dest_conn,
                memlet=memlet,
            )


@dataclasses.dataclass(frozen=True)
class EmptyInputEdge(DataflowInputEdge):
    """
    Allows to setup an edge from a map entry node to a tasklet with no arguments.

    The reason behind this kind of connection is that all nodes inside a map scope
    must have an in/out path that traverses the entry and exit nodes.
    """

    state: dace.SDFGState
    node: dace_nodes.Tasklet

    def connect(self, map_entry: Optional[dace_nodes.MapEntry]) -> None:
        if map_entry is None:
            # outside of a map scope it is possible to instantiate a tasklet node
            # without input connectors
            return
        self.state.add_nedge(map_entry, self.node, dace.Memlet())


@dataclasses.dataclass(frozen=True)
class DataflowOutputEdge:
    """
    Allows to setup an output memlet through a map exit node.

    The result of a dataflow subgraph needs to be written to an external data node.
    The most common case is that the dataflow represents a stencil and the dataflow
    is computed over a field domain, therefore the dataflow is instatiated inside
    a map scope. The `connect` method creates a memlet that writes the dataflow
    result to the external array passing through the `map_exit` node.
    The dataflow can also be instatiated without a map, in which case the `map_exit`
    argument is set to `None`.
    """

    state: dace.SDFGState
    result: ValueExpr

    def connect(
        self,
        map_exit: Optional[dace_nodes.MapExit],
        dest: dace_nodes.AccessNode,
        dest_subset: dace_subsets.Range,
    ) -> bool:
        """Create a connection to the `dest` node, writing the given `dest_subset`.

        Might remove the last data container. Returns `True` if it was removed,
        `False` if kept. For edges writing data from a nested SDFG, in case the
        outside data container is removed, the caller is responsible to propagate
        the strides of the destination array to the array inside the nested SDFG.
        """
        dest_desc = self.result.dc_node.desc(self.state)
        write_edge = self.state.in_edges(self.result.dc_node)[0]

        # Check the kind of node which writes the result
        if isinstance(write_edge.src, dace_nodes.Tasklet):
            # The temporary data written by a tasklet can be safely deleted.
            assert map_exit is not None
            remove_last_node = True
        elif isinstance(write_edge.src, dace_nodes.NestedSDFG):
            if isinstance(dest_desc, dace.data.Scalar):
                # We keep scalar temporary storage, as a general rule, since it
                # does not affect performance of the generated code. This scalar
                # is only required in some cases, e.g. for nested SDFGs implementing
                # reduction, which use a WCR memlet for the reduction operation.
                remove_last_node = False
            else:
                # We remove the transient array on the output connection of a nested
                # SDFG and write directly to the destination node.
                # The caller is responsible to propagate the strides of the destination
                # array to the array inside the nested SDFG.
                remove_last_node = True
        else:
            remove_last_node = False

        if remove_last_node:
            src_node = write_edge.src
            src_node_connector = write_edge.src_conn
            src_subset = write_edge.data.src_subset
            self.state.remove_node(self.result.dc_node)
        else:
            src_node = write_edge.dst
            src_node_connector = None
            src_subset = write_edge.data.dst_subset

        if map_exit is None:
            self.state.add_edge(
                src_node,
                src_node_connector,
                dest,
                None,
                dace.Memlet(data=dest.data, subset=dest_subset, other_subset=src_subset),
            )
        else:
            self.state.add_memlet_path(
                src_node,
                map_exit,
                dest,
                src_conn=src_node_connector,
                memlet=dace.Memlet(data=dest.data, subset=dest_subset, other_subset=src_subset),
            )

        return remove_last_node


DACE_REDUCTION_MAPPING: dict[str, dace.dtypes.ReductionType] = {
    "minimum": dace.dtypes.ReductionType.Min,
    "maximum": dace.dtypes.ReductionType.Max,
    "plus": dace.dtypes.ReductionType.Sum,
    "multiplies": dace.dtypes.ReductionType.Product,
    "and_": dace.dtypes.ReductionType.Logical_And,
    "or_": dace.dtypes.ReductionType.Logical_Or,
    "xor_": dace.dtypes.ReductionType.Logical_Xor,
    "minus": dace.dtypes.ReductionType.Sub,
    "divides": dace.dtypes.ReductionType.Div,
}


def get_reduce_params(node: gtir.FunCall) -> tuple[str, SymbolExpr, SymbolExpr]:
    assert isinstance(node.type, ts.ScalarType)
    dc_dtype = gtx_dace_args.as_dace_type(node.type)

    assert isinstance(node.fun, gtir.FunCall)
    assert len(node.fun.args) == 2
    assert isinstance(node.fun.args[0], gtir.SymRef)
    op_name = str(node.fun.args[0])
    assert isinstance(node.fun.args[1], gtir.Literal)
    assert node.fun.args[1].type == node.type
    reduce_init = SymbolExpr(node.fun.args[1].value, dc_dtype)

    if op_name not in DACE_REDUCTION_MAPPING:
        raise RuntimeError(f"Reduction operation '{op_name}' not supported.")
    identity_value = dace.dtypes.reduction_identity(dc_dtype, DACE_REDUCTION_MAPPING[op_name])
    reduce_identity = SymbolExpr(identity_value, dc_dtype)

    return op_name, reduce_init, reduce_identity


def get_tuple_type(
    data: NestedTuple[IteratorExpr | MemletExpr | ValueExpr],
) -> ts.TupleType:
    """
    Compute the `ts.TupleType` corresponding to the tuple structure of input data expressions.
    """
    data_types: list[ts.DataType] = []
    for dataitem in data:
        if isinstance(dataitem, tuple):
            data_types.append(get_tuple_type(dataitem))
        elif isinstance(dataitem, IteratorExpr):
            data_types.append(dataitem.get_field_type())
        elif isinstance(dataitem, MemletExpr):
            data_types.append(dataitem.gt_dtype)
        else:
            data_types.append(dataitem.gt_dtype)
    return ts.TupleType(data_types)


@dataclasses.dataclass(frozen=True)
class LambdaToDataflow(eve.NodeVisitor):
    """
    Visitor class to translate a `Lambda` expression to a dataflow graph.

    This visitor should be applied by calling `apply()` method on a `Lambda` IR.
    The dataflow graph generated here typically represents the stencil function
    of a field operator. It only computes single elements or pure local fields,
    in case of neighbor values. In case of local fields, the dataflow contains
    inner maps with fixed literal size (max number of neighbors).
    Once the lambda expression has been lowered to a dataflow, the dataflow graph
    needs to be instantiated, that is we have to connect all in/out edges to
    external source/destination data nodes. Since the lambda expression is used
    in GTIR as argument to a field operator, the dataflow is instatiated inside
    a map scope and applied on the field domain. Therefore, all in/out edges
    must traverse the entry/exit map nodes.
    """

    sdfg: dace.SDFG
    state: dace.SDFGState
    subgraph_builder: gtir_to_sdfg.DataflowBuilder
    input_edges: list[DataflowInputEdge] = dataclasses.field(default_factory=lambda: [])
    symbol_map: dict[
        str,
        MaybeNestedInTuple[IteratorExpr | DataExpr],
    ] = dataclasses.field(default_factory=dict)

    def _add_input_data_edge(
        self,
        src: dace_nodes.AccessNode,
        src_subset: dace_subsets.Range,
        dst_node: dace_nodes.Node,
        dst_conn: Optional[str] = None,
    ) -> None:
        edge = MemletInputEdge(self.state, src, src_subset, dst_node, dst_conn)
        self.input_edges.append(edge)

    def _add_edge(
        self,
        src_node: dace.Node,
        src_node_connector: Optional[str],
        dst_node: dace.Node,
        dst_node_connector: Optional[str],
        memlet: dace.Memlet,
    ) -> None:
        """Helper method to add an edge in current state."""
        self.state.add_edge(src_node, src_node_connector, dst_node, dst_node_connector, memlet)

    def _add_map(
        self,
        name: str,
        ndrange: Union[
            Dict[str, Union[str, dace.subsets.Subset]],
            List[Tuple[str, Union[str, dace.subsets.Subset]]],
        ],
        **kwargs: Any,
    ) -> Tuple[dace_nodes.MapEntry, dace_nodes.MapExit]:
        """
        Helper method to add a map in current state.

        The subgraph builder ensures that the map receives a unique name,
        by adding a unique suffix to the provided name.
        """
        return self.subgraph_builder.add_map(name, self.state, ndrange, **kwargs)

    def _add_tasklet(
        self,
        name: str,
        inputs: set[str] | Mapping[str, dace.dtypes.typeclass | None],
        outputs: set[str] | Mapping[str, dace.dtypes.typeclass | None],
        code: str,
        **kwargs: Any,
    ) -> tuple[dace_nodes.Tasklet, dict[str, str]]:
        """
        Helper method to add a tasklet in current state.

        The subgraph builder ensures that the tasklet receives a unique name,
        by adding a unique suffix to the provided name.
        """
        tasklet_node, connector_mapping = self.subgraph_builder.add_tasklet(
            name, self.sdfg, self.state, inputs, outputs, code, **kwargs
        )
        if len(inputs) == 0:
            # All nodes inside a map scope must have an in/out path that traverses
            # the entry and exit nodes. Therefore, a tasklet node with no arguments
            # still needs an (empty) input edge from map entry node.
            edge = EmptyInputEdge(self.state, tasklet_node)
            self.input_edges.append(edge)

        return tasklet_node, connector_mapping

    def _add_mapped_tasklet(
        self,
        name: str,
        map_ranges: Mapping[str, str | dace.subsets.Subset],
        inputs: Mapping[str, dace.Memlet],
        code: str,
        outputs: Mapping[str, dace.Memlet],
        **kwargs: Any,
    ) -> tuple[dace_nodes.Tasklet, dace_nodes.MapEntry, dace_nodes.MapExit, dict[str, str]]:
        """
        Helper method to add a mapped tasklet in current state.

        The subgraph builder ensures that the tasklet receives a unique name,
        by adding a unique suffix to the provided name.
        """
        return self.subgraph_builder.add_mapped_tasklet(
            name, self.sdfg, self.state, map_ranges, inputs, code, outputs, **kwargs
        )

    def _construct_local_view(self, field: MemletExpr | ValueExpr) -> ValueExpr:
        if isinstance(field, MemletExpr):
            desc = field.dc_node.desc(self.sdfg)
            local_dim_indices = [i for i, size in enumerate(field.subset.size()) if size != 1]
            if len(local_dim_indices) == 0:
                # we are accessing a single-element array with shape (1,)
                view_shape = (1,)
                view_strides = (1,)
            else:
                view_shape = tuple(desc.shape[i] for i in local_dim_indices)
                view_strides = tuple(desc.strides[i] for i in local_dim_indices)
            view, _ = self.sdfg.add_view(
                f"view_{field.dc_node.data}",
                view_shape,
                desc.dtype,
                strides=view_strides,
                find_new_name=True,
            )
            local_view_node = self.state.add_access(view)
            self._add_input_data_edge(field.dc_node, field.subset, local_view_node)

            return ValueExpr(local_view_node, field.gt_dtype)

        else:
            return field

    def _construct_tasklet_result(
        self,
        dc_dtype: dace.typeclass,
        src_node: dace_nodes.Tasklet,
        src_connector: str,
        use_array: bool = False,
    ) -> ValueExpr:
        data_type = gtx_dace_args.as_itir_type(dc_dtype)
        if use_array:
            # In some cases, such as result data with list-type annotation, we want
            # that output data is represented as an array (single-element 1D array)
            # in order to allow for composition of array shape in external memlets.
            temp_name, _ = self.subgraph_builder.add_temp_array(self.sdfg, (1,), dc_dtype)
        else:
            temp_name, _ = self.subgraph_builder.add_temp_scalar(self.sdfg, dc_dtype)

        temp_node = self.state.add_access(temp_name)
        self._add_edge(
            src_node,
            src_connector,
            temp_node,
            None,
            dace.Memlet(data=temp_name, subset="0"),
        )
        return ValueExpr(
            dc_node=temp_node,
            gt_dtype=(
                ts.ListType(element_type=data_type, offset_type=_CONST_DIM)
                if use_array
                else data_type
            ),
        )

    def _visit_deref(self, node: gtir.FunCall) -> DataExpr:
        """
        Visit a `deref` node, which represents dereferencing of an iterator.
        The iterator is the argument of this node.

        The iterator contains the information for accessing a field, that is the
        sorted list of dimensions in the field domain and the index values for
        each dimension. The index values can be either symbol values, that is
        literal values or scalar arguments which are constant in the SDFG scope;
        or they can be the result of some expression, that computes a dynamic
        index offset or gets an neighbor index from a connectivity table.
        In case all indexes are symbol values, the `deref` node is lowered to a
        memlet; otherwise dereferencing is a runtime operation represented in
        the SDFG as a tasklet node.
        """
        # format used for field index tasklet connector
        IndexConnectorFmt: Final = "index_{dim}"

        if isinstance(node.type, ts.TupleType):
            raise NotImplementedError("Tuple deref not supported.")

        assert len(node.args) == 1
        arg_expr = self.visit(node.args[0])

        if not isinstance(arg_expr, IteratorExpr):
            # dereferencing a scalar or a literal node results in the node itself
            return arg_expr

        field_desc = arg_expr.field.desc(self.sdfg)
        if isinstance(field_desc, dace.data.Scalar):
            # deref a zero-dimensional field
            assert len(arg_expr.field_domain) == 0
            assert isinstance(node.type, ts.ScalarType)
            return MemletExpr(
                dc_node=arg_expr.field,
                gt_field=ts.FieldType(dims=[], dtype=arg_expr.gt_dtype),
                subset="0",
            )

        # handle default case below: deref a field with one or more dimensions

        # when the indices are all dace symbolic expressions, the deref is lowered
        # to a memlet, where the index is the memlet subset
        if all(isinstance(index, SymbolExpr) for index in arg_expr.indices.values()):
            # when all indices are symbolic expressions, we can perform direct field access through a memlet
            field_subset = arg_expr.get_memlet_subset(self.sdfg)
            return MemletExpr(arg_expr.field, arg_expr.get_field_type(), field_subset)

        # when any of the indices is a runtime value (either a dynamic cartesian
        # offset or a connectivity offset), the deref is lowered to a tasklet
        assert all(dim in arg_expr.indices for dim, _ in arg_expr.field_domain)
        assert len(field_desc.shape) == len(arg_expr.field_domain)
        field_indices = [(dim, arg_expr.indices[dim]) for dim, _ in arg_expr.field_domain]
        index_connectors = [
            IndexConnectorFmt.format(dim=dim.value)
            for dim, index in field_indices
            if not isinstance(index, SymbolExpr)
        ]
        # here `internals` refer to the names used as index in the tasklet code string:
        # an index can be either a connector name (for dynamic/indirect indices)
        # or a symbol value (for literal values and scalar arguments).
        index_internals = ",".join(
            str(index.value - offset)
            if isinstance(index := arg_expr.indices[dim], SymbolExpr)
            else f"{IndexConnectorFmt.format(dim=dim.value)} - {offset}"
            for (dim, offset) in arg_expr.field_domain
        )
        deref_node, connector_mapping = self._add_tasklet(
            name="deref",
            inputs={"field"} | set(index_connectors),
            outputs={"val"},
            code=f"val = field[{index_internals}]",
        )
        # add new termination point for the field parameter
        self._add_input_data_edge(
            arg_expr.field,
            dace_subsets.Range.from_array(field_desc),
            deref_node,
            connector_mapping["field"],
        )

        # add termination points for the dynamic iterator indices
        for dim, index_expr in field_indices:
            index_connector = IndexConnectorFmt.format(dim=dim.value)
            if isinstance(index_expr, MemletExpr):
                self._add_input_data_edge(
                    index_expr.dc_node,
                    index_expr.subset,
                    deref_node,
                    connector_mapping[index_connector],
                )

            elif isinstance(index_expr, ValueExpr):
                self._add_edge(
                    index_expr.dc_node,
                    None,
                    deref_node,
                    connector_mapping[index_connector],
                    dace.Memlet(data=index_expr.dc_node.data, subset="0"),
                )
            else:
                assert isinstance(index_expr, SymbolExpr)

        return self._construct_tasklet_result(
            field_desc.dtype, deref_node, connector_mapping["val"]
        )

    def _visit_if_branch_arg(
        self,
        if_sdfg: dace.SDFG,
        if_branch_state: dace.SDFGState,
        param_name: str,
        arg: IteratorExpr | DataExpr,
        deref_on_input_memlet: bool,
        if_sdfg_input_memlets: dict[str, MemletExpr | ValueExpr],
    ) -> IteratorExpr | ValueExpr:
        """
        Helper method to be called by `_visit_if_branch()` to visit the input arguments.

        Args:
            if_sdfg: The nested SDFG where the if expression is lowered.
            if_branch_state: The state inside the nested SDFG where the if branch is lowered.
            param_name: The parameter name of the input argument.
            arg: The input argument expression.
            deref_on_input_memlet: When True, the given iterator argument can be dereferenced on the input memlet.
            if_sdfg_input_memlets: The memlets that provide input data to the nested SDFG, will be update inside this function.
        """
        use_full_shape = False
        if isinstance(arg, (MemletExpr, ValueExpr)):
            arg_desc = arg.dc_node.desc(self.sdfg)
            arg_expr = arg
        elif isinstance(arg, IteratorExpr):
            arg_desc = arg.field.desc(self.sdfg)
            if deref_on_input_memlet:
                # If the iterator is just dereferenced inside the branch state,
                # we can access the array outside the nested SDFG and pass the
                # local data. This approach makes the data dependencies of nested
                # structures more explicit and thus makes it easier for MapFusion
                # to correctly infer the data dependencies.
                memlet_subset = arg.get_memlet_subset(self.sdfg)
                arg_expr = MemletExpr(arg.field, arg.get_field_type(), memlet_subset)
            else:
                # In order to shift the iterator inside the branch dataflow,
                # we have to pass the full array shape.
                arg_expr = MemletExpr(
                    arg.field, arg.get_field_type(), dace_subsets.Range.from_array(arg_desc)
                )
                use_full_shape = True
        else:
            raise TypeError(f"Unexpected {arg} as input argument.")

        if use_full_shape:
            inner_desc = arg_desc.clone()
            inner_desc.transient = False
        elif isinstance(arg.gt_dtype, ts.ScalarType):
            inner_desc = dace.data.Scalar(arg_desc.dtype)
        else:
            # for list of values, we retrieve the local size from the corresponding offset
            assert arg.gt_dtype.offset_type is not None
            offset_provider_type = self.subgraph_builder.get_offset_provider_type(
                arg.gt_dtype.offset_type.value
            )
            assert isinstance(offset_provider_type, gtx_common.NeighborConnectivityType)
            inner_desc = dace.data.Array(
                dtype=arg_desc.dtype, shape=[offset_provider_type.max_neighbors]
            )

        if param_name in if_sdfg.arrays:
            # the data desciptor was added by the visitor of the other branch expression
            assert if_sdfg.data(param_name) == inner_desc
        else:
            if_sdfg.add_datadesc(param_name, inner_desc)
            if_sdfg_input_memlets[param_name] = arg_expr

        inner_node = if_branch_state.add_access(param_name)
        if isinstance(arg, IteratorExpr) and use_full_shape:
            return IteratorExpr(inner_node, arg.gt_dtype, arg.field_domain, arg.indices)
        else:
            return ValueExpr(inner_node, arg.gt_dtype)

    def _visit_if_branch(
        self,
        if_sdfg: dace.SDFG,
        if_branch_state: dace.SDFGState,
        expr: gtir.Expr,
        if_sdfg_input_memlets: dict[str, MemletExpr | ValueExpr],
        direct_deref_iterators: Iterable[str],
    ) -> tuple[list[DataflowInputEdge], MaybeNestedInTuple[DataflowOutputEdge]]:
        """
        Helper method to visit an if-branch expression and lower it to a dataflow inside the given nested SDFG and state.

        This function is called by `_visit_if()` for each if-branch.

        Args:
            if_sdfg: The nested SDFG where the if expression is lowered.
            if_branch_state: The state inside the nested SDFG where the if branch is lowered.
            expr: The if branch expression to lower.
            if_sdfg_input_memlets: The memlets that provide input data to the nested SDFG, will be update inside this function.
            direct_deref_iterators: Fields that are accessed with direct iterator deref, without any shift.

        Returns:
            A tuple containing:
                - the list of input edges for the parent dataflow
                - the tree representation of output data, in the form of a tuple of data edges.
        """
        assert if_branch_state in if_sdfg.states()

        lambda_args = []
        lambda_params = []
        for pname in symbol_ref_utils.collect_symbol_refs(expr, self.symbol_map.keys()):
            arg = self.symbol_map[pname]
            if isinstance(arg, tuple):
                ptype = get_tuple_type(arg)  # type: ignore[arg-type]
                psymbol = im.sym(pname, ptype)
                psymbol_tree = gtir_to_sdfg_utils.make_symbol_tree(pname, ptype)
                deref_on_input_memlet = pname in direct_deref_iterators
                inner_arg = gtx_utils.tree_map(
                    lambda tsym,
                    targ,
                    deref_on_input_memlet=deref_on_input_memlet: self._visit_if_branch_arg(
                        if_sdfg,
                        if_branch_state,
                        str(tsym.id),
                        targ,
                        deref_on_input_memlet,
                        if_sdfg_input_memlets,
                    )
                )(psymbol_tree, arg)
            else:
                psymbol = im.sym(pname, arg.gt_dtype)  # type: ignore[union-attr]
                deref_on_input_memlet = pname in direct_deref_iterators
                inner_arg = self._visit_if_branch_arg(
                    if_sdfg,
                    if_branch_state,
                    pname,
                    arg,
                    deref_on_input_memlet,
                    if_sdfg_input_memlets,
                )
            lambda_args.append(inner_arg)
            lambda_params.append(psymbol)

        # visit each branch of the if-statement as if it was a Lambda node
        lambda_node = gtir.Lambda(params=lambda_params, expr=expr)
        input_edges, output_tree = translate_lambda_to_dataflow(
            if_sdfg, if_branch_state, self.subgraph_builder, lambda_node, lambda_args
        )

        return input_edges, output_tree

    def _visit_if_branch_result(
        self, sdfg: dace.SDFG, state: dace.SDFGState, edge: DataflowOutputEdge, sym: gtir.Sym
    ) -> ValueExpr:
        """
        Helper function to be called by `_visit_if` to create an output connector
        on the nested SDFG that will write the result to the parent SDFG.
        The result data inside the nested SDFG must have the same name as the connector.
        """
        output_data = str(sym.id)
        if output_data in sdfg.arrays:
            output_desc = sdfg.data(output_data)
            assert not output_desc.transient
        else:
            # If the result is currently written to a transient node, inside the nested SDFG,
            # we need to allocate a non-transient data node.
            result_desc = edge.result.dc_node.desc(sdfg)
            output_desc = result_desc.clone()
            output_desc.transient = False
            output_data = sdfg.add_datadesc(output_data, output_desc, find_new_name=True)
        output_node = state.add_access(output_data)
        state.add_nedge(
            edge.result.dc_node,
            output_node,
            dace.Memlet.from_array(output_data, output_desc),
        )
        return ValueExpr(output_node, edge.result.gt_dtype)

    def _visit_if(self, node: gtir.FunCall) -> MaybeNestedInTuple[ValueExpr]:
        """
        Lowers an if-expression with exclusive branch execution into a nested SDFG,
        in which each branch is lowered into a dataflow in a separate state and
        the if-condition is represented as the inter-state edge condition.
        """

        def write_output_of_nested_sdfg_to_temporary(inner_value: ValueExpr) -> ValueExpr:
            # Each output connector of the nested SDFG writes to a transient node in the parent SDFG
            inner_data = inner_value.dc_node.data
            inner_desc = inner_value.dc_node.desc(nsdfg)
            assert not inner_desc.transient
            output, output_desc = self.subgraph_builder.add_temp_array_like(self.sdfg, inner_desc)
            output_node = self.state.add_access(output)
            self.state.add_edge(
                nsdfg_node,
                inner_data,
                output_node,
                None,
                dace.Memlet.from_array(output, output_desc),
            )
            return ValueExpr(output_node, inner_value.gt_dtype)

        assert len(node.args) == 3

        # evaluate the if-condition that will write to a boolean scalar node
        condition_value = self.visit(node.args[0])
        assert (
            (
                isinstance(condition_value.gt_dtype, ts.ScalarType)
                and condition_value.gt_dtype.kind == ts.ScalarKind.BOOL
            )
            if isinstance(condition_value, (MemletExpr, ValueExpr))
            else (condition_value.dc_dtype == dace.dtypes.bool_)
        )

        nsdfg = dace.SDFG(self.subgraph_builder.unique_nsdfg_name("if_stmt"))
        nsdfg.debuginfo = gtir_to_sdfg_utils.debug_info(node, default=self.sdfg.debuginfo)

        # create states inside the nested SDFG for the if-branches
        if_region = dace.sdfg.state.ConditionalBlock("if")
        nsdfg.add_node(if_region, ensure_unique_name=True)
        entry_state = nsdfg.add_state("entry", is_start_block=True)
        nsdfg.add_edge(entry_state, if_region, dace.InterstateEdge())

        then_body = dace.sdfg.state.ControlFlowRegion("then_body", sdfg=nsdfg)
        tstate = then_body.add_state("true_branch", is_start_block=True)
        if_region.add_branch(dace.sdfg.state.CodeBlock("__cond"), then_body)

        else_body = dace.sdfg.state.ControlFlowRegion("else_body", sdfg=nsdfg)
        fstate = else_body.add_state("false_branch", is_start_block=True)
        # Use `None` for unconditional execution of else-branch, if the condition is not met.
        if_region.add_branch(None, else_body)

        input_memlets: dict[str, MemletExpr | ValueExpr] = {}
        nsdfg_symbols_mapping: Optional[dict[str, dace.symbol]] = None

        # define scalar or symbol for the condition value inside the nested SDFG
        if isinstance(condition_value, SymbolExpr):
            nsdfg.add_symbol("__cond", dace.dtypes.bool)
        else:
            nsdfg.add_scalar("__cond", dace.dtypes.bool)
            input_memlets["__cond"] = condition_value

        # Collect all field iterators that are shifted inside any of the then/else
        # branch expressions. Iterator shift expressions require the field argument
        # as iterator, therefore the corresponding array has to be passed with full
        # shape into the nested SDFG where the if_ expression is lowered. When the
        # branch expression simply does `deref` on the iterator, without any shifting,
        # it corresponds to a direct element access. Such `deref` expressions can
        # be lowered outside the nested SDFG, so that just the local value (a scalar
        # or a list of values) is passed as input to the nested SDFG.
        shifted_iterator_symbols = set()
        for branch_expr in node.args[1:3]:
            for shift_node in eve.walk_values(branch_expr).filter(
                lambda x: cpm.is_applied_shift(x)
            ):
                shifted_iterator_symbols |= (
                    eve.walk_values(shift_node)
                    .if_isinstance(gtir.SymRef)
                    .map(lambda x: str(x.id))
                    .filter(lambda x: isinstance(self.symbol_map.get(x, None), IteratorExpr))
                    .to_set()
                )
        iterator_symbols = {
            sym_name
            for sym_name, sym_type in self.symbol_map.items()
            if isinstance(sym_type, IteratorExpr)
        }
        direct_deref_iterators = (
            set(symbol_ref_utils.collect_symbol_refs(node.args[1:3], iterator_symbols))
            - shifted_iterator_symbols
        )

        for nstate, arg in zip([tstate, fstate], node.args[1:3]):
            # visit each if-branch in the corresponding state of the nested SDFG
            in_edges, out_edges = self._visit_if_branch(
                nsdfg, nstate, arg, input_memlets, direct_deref_iterators
            )
            for edge in in_edges:
                edge.connect(map_entry=None)

            out_syms = (
                gtir_to_sdfg_utils.make_symbol_tree("__output", node.type)
                if isinstance(node.type, ts.TupleType)
                else im.sym("__output", node.type)
            )

            outer_value = gtx_utils.tree_map(
                lambda edge, sym, nstate_=nstate: self._visit_if_branch_result(
                    nsdfg, nstate_, edge, sym
                )
            )(out_edges, out_syms)

        else:
            result = outer_value

        outputs = {outval.dc_node.data for outval in gtx_utils.flatten_nested_tuple((result,))}

        # all free symbols are mapped to the symbols available in parent SDFG
        nsdfg_symbols_mapping = {str(sym): sym for sym in nsdfg.free_symbols}
        if isinstance(condition_value, SymbolExpr):
            nsdfg_symbols_mapping["__cond"] = condition_value.value
        nsdfg_node = self.state.add_nested_sdfg(
            nsdfg,
            inputs=set(input_memlets.keys()),
            outputs=outputs,
            symbol_mapping=nsdfg_symbols_mapping,
        )

        for inner, input_expr in input_memlets.items():
            if isinstance(input_expr, MemletExpr):
                self._add_input_data_edge(input_expr.dc_node, input_expr.subset, nsdfg_node, inner)
            else:
                self._add_edge(
                    input_expr.dc_node,
                    None,
                    nsdfg_node,
                    inner,
                    self.sdfg.make_array_memlet(input_expr.dc_node.data),
                )

        return (
            gtx_utils.tree_map(write_output_of_nested_sdfg_to_temporary)(result)
            if isinstance(result, tuple)
            else write_output_of_nested_sdfg_to_temporary(result)
        )

    def _visit_neighbors(self, node: gtir.FunCall) -> ValueExpr:
        assert isinstance(node.type, ts.ListType)
        assert len(node.args) == 2

        assert isinstance(node.args[0], gtir.OffsetLiteral)
        offset = node.args[0].value
        assert isinstance(offset, str)
        conn_type = self.subgraph_builder.get_offset_provider_type(offset)
        assert isinstance(conn_type, gtx_common.NeighborConnectivityType)

        it = self.visit(node.args[1])
        assert isinstance(it, IteratorExpr)
        assert any(dim == conn_type.codomain for dim, _ in it.field_domain)
        assert conn_type.source_dim in it.indices
        origin_index = it.indices[conn_type.source_dim]
        assert isinstance(origin_index, SymbolExpr)
        assert all(isinstance(index, SymbolExpr) for index in it.indices.values())

        # initially, the storage for the connectivty tables is created as transient;
        # when the tables are used, the storage is changed to non-transient,
        # as the corresponding arrays are supposed to be allocated by the SDFG caller
        conn_data = gtx_dace_args.connectivity_identifier(offset)
        conn_desc = self.sdfg.arrays[conn_data]
        conn_desc.transient = False

        # The visitor is constructing a list of input connections that will be handled
        # by `translate_as_fieldop` (the primitive translator), that is responsible
        # of creating the map for the field domain. For each input connection, it will
        # create a memlet that will write to a node specified by the third attribute
        # in the `InputConnection` tuple (either a tasklet, or a view node, or a library
        # node). For the specific case of `neighbors` we need to nest the neighbors map
        # inside the field map and the memlets will traverse the external map and write
        # to the view nodes. The simplify pass will remove the redundant access nodes.
        field_desc = it.field.desc(self.sdfg)
        field_slice = self._construct_local_view(
            MemletExpr(
                dc_node=it.field,
                gt_field=it.get_field_type(),
                subset=dace_subsets.Range.from_string(
                    ",".join(
                        str(it.indices[dim].value - offset)  # type: ignore[union-attr]
                        if dim != conn_type.codomain
                        else f"0:{size}"
                        for (dim, offset), size in zip(
                            it.field_domain, field_desc.shape, strict=True
                        )
                    )
                ),
            )
        )
        # The layout of connectivity tables is known.
        assert len(conn_type.domain) == 2
        assert conn_type.domain[1].kind == gtx_common.DimensionKind.LOCAL
        conn_slice = self._construct_local_view(
            MemletExpr(
                dc_node=self.state.add_access(conn_data),
                gt_field=ts.FieldType(
                    dims=[conn_type.domain[0]],
                    dtype=ts.ListType(
                        element_type=tt.from_dtype(conn_type.dtype), offset_type=_CONST_DIM
                    ),
                ),
                subset=dace_subsets.Range.from_string(
                    f"{origin_index.value}, 0:{conn_type.max_neighbors}"
                ),
            )
        )

        neighbors_temp, _ = self.subgraph_builder.add_temp_array(
            self.sdfg, (conn_type.max_neighbors,), field_desc.dtype
        )
        neighbors_node = self.state.add_access(neighbors_temp)
        offset_type = gtx_common.Dimension(offset, gtx_common.DimensionKind.LOCAL)
        neighbor_idx = gtir_to_sdfg_utils.get_map_variable(offset_type)

        index_connector = "__index"
        field_connector = "__field"
        output_connector = "__val"
        tasklet_expression = f"{output_connector} = {field_connector}[{index_connector}]"
        input_memlets = {
            field_connector: self.sdfg.make_array_memlet(field_slice.dc_node.data),
            index_connector: dace.Memlet(data=conn_slice.dc_node.data, subset=neighbor_idx),
        }
        input_nodes = {
            field_slice.dc_node.data: field_slice.dc_node,
            conn_slice.dc_node.data: conn_slice.dc_node,
        }

        if conn_type.has_skip_values:
            # in case of skip value we can write any dummy value
            skip_value = (
                "math.nan"
                if ti.is_floating_point(node.type.element_type)
                else str(dace.dtypes.max_value(field_desc.dtype))
            )
            tasklet_expression += (
                f" if {index_connector} != {gtx_common._DEFAULT_SKIP_VALUE} else {skip_value}"
            )

        self._add_mapped_tasklet(
            name=f"{offset}_neighbors",
            map_ranges={neighbor_idx: f"0:{conn_type.max_neighbors}"},
            code=tasklet_expression,
            inputs=input_memlets,
            input_nodes=input_nodes,
            outputs={
                output_connector: dace.Memlet(data=neighbors_temp, subset=neighbor_idx),
            },
            output_nodes={neighbors_temp: neighbors_node},
            external_edges=True,
        )

        return ValueExpr(
            dc_node=neighbors_node, gt_dtype=ts.ListType(node.type.element_type, offset_type)
        )

    def _visit_list_get(self, node: gtir.FunCall) -> ValueExpr:
        assert len(node.args) == 2
        index_arg = self.visit(node.args[0])
        src_arg = self.visit(node.args[1])
        assert isinstance(src_arg.gt_dtype, ts.ListType)

        src_desc = src_arg.dc_node.desc(self.sdfg)
        if isinstance(src_arg, MemletExpr):
            assert len(src_desc.shape) == len(src_arg.subset)
            src_subset = src_arg.subset
            local_dim = src_arg.gt_dtype.offset_type
            all_dims = gtx_common.order_dimensions([*src_arg.gt_field.dims, local_dim])  # type: ignore[list-item]  # checked in MemletExpr.__post_init__
            local_dim_index = all_dims.index(local_dim)  # type: ignore[arg-type]  # checked in MemletExpr.__post_init__
        elif isinstance(src_arg, ValueExpr):
            assert len(src_desc.shape) == 1
            src_subset = dace_subsets.Range.from_array(src_desc)
            local_dim_index = 0
        else:
            raise ValueError(f"Unexpected argument type {type(src_arg)} in 'list_get' expression.")

        assert isinstance(src_arg.gt_dtype.element_type, ts.ScalarType)
        assert src_desc.dtype == gtx_dace_args.as_dace_type(src_arg.gt_dtype.element_type)
        dst, _ = self.subgraph_builder.add_temp_scalar(self.sdfg, src_desc.dtype)
        dst_node = self.state.add_access(dst)

        if isinstance(index_arg, SymbolExpr):
            assert index_arg.dc_dtype in dace.dtypes.INTEGER_TYPES
            src_subset = (
                dace_subsets.Range(src_subset[:local_dim_index])
                + dace_subsets.Range.from_string(index_arg.value)
                + dace_subsets.Range(src_subset[local_dim_index + 1 :])
            )
            if isinstance(src_arg, MemletExpr):
                self._add_input_data_edge(src_arg.dc_node, src_subset, dst_node)
            else:
                self._add_edge(
                    src_arg.dc_node,
                    None,
                    dst_node,
                    None,
                    dace.Memlet(data=src_arg.dc_node.data, subset=src_subset, other_subset="0"),
                )
        elif isinstance(index_arg, ValueExpr):
            tasklet_node, connector_mapping = self._add_tasklet(
                name="list_get",
                inputs={"index", "data"},
                outputs={"val"},
                code="val = data[index]",
            )
            self._add_edge(
                index_arg.dc_node,
                None,
                tasklet_node,
                connector_mapping["index"],
                dace.Memlet(data=index_arg.dc_node.data, subset="0"),
            )
            if isinstance(src_arg, MemletExpr):
                self._add_input_data_edge(
                    src_arg.dc_node, src_subset, tasklet_node, connector_mapping["data"]
                )
            else:
                self._add_edge(
                    src_arg.dc_node, None, tasklet_node, connector_mapping["data"], src_subset
                )
            self._add_edge(
                tasklet_node,
                connector_mapping["val"],
                dst_node,
                None,
                dace.Memlet(data=dst, subset="0"),
            )
        else:
            raise TypeError(f"Unexpected value {index_arg} as index argument.")

        return ValueExpr(dc_node=dst_node, gt_dtype=src_arg.gt_dtype.element_type)

    def _visit_map(self, node: gtir.FunCall) -> ValueExpr:
        """
        A map node defines an operation to be mapped on all elements of input arguments.

        The map operation is applied on the local dimension of input fields.
        In the example below, the local dimension consists of a list of neighbor
        values as the first argument, and a list of constant values `1.0`:
        `map_(plus)(neighbors(V2E, it), make_const_list(1.0))`

        The `plus` operation is lowered to a tasklet inside a map that computes
        the domain of the local dimension (in this example, max neighbors in V2E).

        The result is a 1D local field, with same size as the input local dimension.
        In above example, the result would be an array with size V2E.max_neighbors,
        containing the V2E neighbor values incremented by 1.0.
        """
        assert isinstance(node.type, ts.ListType)
        assert isinstance(node.fun, gtir.FunCall)
        assert len(node.fun.args) == 1  # the operation to be mapped on the arguments

        assert isinstance(node.type.element_type, ts.ScalarType)
        dc_dtype = gtx_dace_args.as_dace_type(node.type.element_type)

        input_connectors = [f"__arg{i}" for i in range(len(node.args))]
        output_connector = "__out"

        # Here we build the body of the tasklet
        fun_node = im.call(node.fun.args[0])(*input_connectors)
        fun_python_code = gtir_python_codegen.get_source(fun_node)
        tasklet_expression = f"{output_connector} = {fun_python_code}"

        input_args = [self.visit(arg) for arg in node.args]
        input_conn_types: dict[gtx_common.Dimension, gtx_common.NeighborConnectivityType] = {}
        for input_arg in input_args:
            assert isinstance(input_arg.gt_dtype, ts.ListType)
            assert input_arg.gt_dtype.offset_type is not None
            offset_type = input_arg.gt_dtype.offset_type
            if offset_type == _CONST_DIM:
                # this input argument is the result of `make_const_list`
                continue
            offset_provider_t = self.subgraph_builder.get_offset_provider_type(offset_type.value)
            assert isinstance(offset_provider_t, gtx_common.NeighborConnectivityType)
            input_conn_types[offset_type] = offset_provider_t

        if len(input_conn_types) == 0:
            raise ValueError(f"Missing information on local dimension for map node {node}.")

        # GT4Py guarantees that all connectivities used to generate lists of neighbors
        # have the same length, that is the same value of 'max_neighbors'.
        if (
            len(
                set(
                    (conn.has_skip_values, conn.max_neighbors) for conn in input_conn_types.values()
                )
            )
            != 1
        ):
            raise ValueError("Unexpected arguments to map expression with different neighborhood.")
        offset_type, conn_type = next(iter(input_conn_types.items()))
        local_size = conn_type.max_neighbors
        map_index = gtir_to_sdfg_utils.get_map_variable(offset_type)

        # The dataflow we build in this class has some loose connections on input edges.
        # These edges are described as set of nodes, that will have to be connected to
        # external data source nodes passing through the map entry node of the field map.
        # Similarly to `neighbors` expressions, the `map_` input edges terminate on view
        # nodes (see `_construct_local_view` in the for-loop below), because it is simpler
        # than representing map-to-map edges (which require memlets with 2 pass-nodes).
        input_memlets = {}
        input_nodes = {}
        for conn, input_arg in zip(input_connectors, input_args):
            input_node = self._construct_local_view(input_arg).dc_node
            input_desc = input_node.desc(self.sdfg)
            # we assume that there is a single local dimension
            if len(input_desc.shape) != 1:
                raise ValueError(f"More than one local dimension in map expression {node}.")
            input_size = input_desc.shape[0]
            if input_size == 1:
                assert input_arg.gt_dtype.offset_type == _CONST_DIM
                input_memlets[conn] = dace.Memlet(data=input_node.data, subset="0")
            elif input_size == local_size:
                input_memlets[conn] = dace.Memlet(data=input_node.data, subset=map_index)
            else:
                raise ValueError(
                    f"Argument to map node with local size {input_size}, expected {local_size}."
                )
            input_nodes[input_node.data] = input_node

        result, _ = self.subgraph_builder.add_temp_array(self.sdfg, (local_size,), dc_dtype)
        result_node = self.state.add_access(result)

        if conn_type.has_skip_values:
            # In case the `map_` input expressions contain skip values, we use
            # the connectivity-based offset provider as mask for map computation.
            conn_data = gtx_dace_args.connectivity_identifier(offset_type.value)
            conn_desc = self.sdfg.arrays[conn_data]
            conn_desc.transient = False

            origin_map_index = gtir_to_sdfg_utils.get_map_variable(conn_type.source_dim)

            # The layout of connectivity tables is known.
            assert len(conn_type.domain) == 2
            assert conn_type.domain[1].kind == gtx_common.DimensionKind.LOCAL
            conn_slice = self._construct_local_view(
                MemletExpr(
                    dc_node=self.state.add_access(conn_data),
                    gt_field=ts.FieldType(
                        dims=[conn_type.domain[0]],
                        dtype=ts.ListType(
                            element_type=tt.from_dtype(conn_type.dtype), offset_type=_CONST_DIM
                        ),
                    ),
                    subset=dace_subsets.Range.from_string(
                        f"{origin_map_index}, 0:{conn_type.max_neighbors}"
                    ),
                )
            )

            input_memlets["__neighbor_idx"] = dace.Memlet(
                data=conn_slice.dc_node.data, subset=map_index
            )
            input_nodes[conn_slice.dc_node.data] = conn_slice.dc_node

            # in case of skip value we can write any dummy value
            skip_value = (
                "math.nan"
                if ti.is_floating_point(node.type.element_type)
                else str(dace.dtypes.max_value(dc_dtype))
            )
            tasklet_expression += (
                f" if __neighbor_idx != {gtx_common._DEFAULT_SKIP_VALUE} else {skip_value}"
            )

        self._add_mapped_tasklet(
            name="map",
            map_ranges={map_index: f"0:{local_size}"},
            code=tasklet_expression,
            inputs=input_memlets,
            input_nodes=input_nodes,
            outputs={
                output_connector: dace.Memlet(data=result, subset=map_index),
            },
            output_nodes={result: result_node},
            external_edges=True,
        )

        return ValueExpr(
            dc_node=result_node,
            gt_dtype=ts.ListType(node.type.element_type, offset_type),
        )

    def _make_reduce_with_skip_values(
        self,
        input_expr: ValueExpr | MemletExpr,
        offset_provider_type: gtx_common.NeighborConnectivityType,
        reduce_init: SymbolExpr,
        reduce_identity: SymbolExpr,
        reduce_wcr: str,
        result_node: dace_nodes.AccessNode,
    ) -> None:
        """
        Helper method to lower reduction on a local field containing skip values.

        The reduction is implemented as a nested SDFG containing 2 states. In first
        state, the result (a scalar data node passed as argumet) is initialized.
        In second state, a mapped tasklet uses a write-conflict resolution (wcr)
        memlet to update the result.
        We use the offset provider as a mask to identify skip values: the value
        that is written to the result node is either the input value, when the
        corresponding neighbor index in the connectivity table is valid, or the
        identity value if the neighbor index is missing.
        """
        origin_map_index = gtir_to_sdfg_utils.get_map_variable(offset_provider_type.source_dim)

        assert (
            isinstance(input_expr.gt_dtype, ts.ListType)
            and input_expr.gt_dtype.offset_type is not None
        )
        offset_type = input_expr.gt_dtype.offset_type
        connectivity = gtx_dace_args.connectivity_identifier(offset_type.value)
        connectivity_node = self.state.add_access(connectivity)
        connectivity_desc = connectivity_node.desc(self.sdfg)
        connectivity_desc.transient = False

        desc = input_expr.dc_node.desc(self.sdfg)
        if isinstance(input_expr, MemletExpr):
            local_dim_indices = [i for i, size in enumerate(input_expr.subset.size()) if size != 1]
        else:
            local_dim_indices = list(range(len(desc.shape)))

        if len(local_dim_indices) != 1:
            raise NotImplementedError(
                f"Found {len(local_dim_indices)} local dimensions in reduce expression, expected one."
            )
        local_dim_index = local_dim_indices[0]
        assert desc.shape[local_dim_index] == offset_provider_type.max_neighbors

        # we lower the reduction map with WCR out memlet in a nested SDFG
        nsdfg = dace.SDFG(self.subgraph_builder.unique_nsdfg_name("reduce_with_skip_values"))
        nsdfg.add_array(
            "values",
            (desc.shape[local_dim_index],),
            desc.dtype,
            strides=(desc.strides[local_dim_index],),
        )
        nsdfg.add_array(
            "neighbor_indices",
            (connectivity_desc.shape[1],),
            connectivity_desc.dtype,
            strides=(connectivity_desc.strides[1],),
        )
        nsdfg.add_scalar("acc", desc.dtype)
        st_init = nsdfg.add_state(f"{nsdfg.label}_init")
        init_tasklet, connector_mapping = self.subgraph_builder.add_tasklet(
            name="init_acc",
            sdfg=self.sdfg,
            state=st_init,
            inputs={},
            outputs={"val"},
            code=f"val = {reduce_init.dc_dtype}({reduce_init.value})",
        )
        st_init.add_edge(
            init_tasklet,
            connector_mapping["val"],
            st_init.add_access("acc"),
            None,
            dace.Memlet(data="acc", subset="0"),
        )
        st_reduce = nsdfg.add_state_after(st_init, f"{nsdfg.label}_reduce")
        # Fill skip values in local dimension with the reduce identity value
        skip_value = f"{reduce_identity.dc_dtype}({reduce_identity.value})"
        # Since this map operates on a pure local dimension, we explicitly set sequential
        # schedule and we set the flag 'wcr_nonatomic=True' on the write memlet.
        # TODO(phimuell): decide if auto-optimizer should reset `wcr_nonatomic` properties, as DaCe does.
        self.subgraph_builder.add_mapped_tasklet(
            name="reduce_with_skip_values",
            sdfg=self.sdfg,
            state=st_reduce,
            map_ranges={"i": f"0:{offset_provider_type.max_neighbors}"},
            inputs={
                "val": dace.Memlet(data="values", subset="i"),
                "neighbor_idx": dace.Memlet(data="neighbor_indices", subset="i"),
            },
            code=f"out = val if neighbor_idx != {gtx_common._DEFAULT_SKIP_VALUE} else {skip_value}",
            outputs={
                "out": dace.Memlet(data="acc", subset="0", wcr=reduce_wcr, wcr_nonatomic=True),
            },
            external_edges=True,
            schedule=dace.dtypes.ScheduleType.Sequential,
        )

        nsdfg_node = self.state.add_nested_sdfg(
            nsdfg, inputs={"values", "neighbor_indices"}, outputs={"acc"}
        )

        if isinstance(input_expr, MemletExpr):
            self._add_input_data_edge(input_expr.dc_node, input_expr.subset, nsdfg_node, "values")
        else:
            self.state.add_edge(
                input_expr.dc_node,
                None,
                nsdfg_node,
                "values",
                self.sdfg.make_array_memlet(input_expr.dc_node.data),
            )
        # The layout of connectivity tables is known.
        assert len(offset_provider_type.domain) == 2
        assert offset_provider_type.domain[1].kind == gtx_common.DimensionKind.LOCAL
        self._add_input_data_edge(
            connectivity_node,
            dace_subsets.Range.from_string(
                f"{origin_map_index}, 0:{offset_provider_type.max_neighbors}"
            ),
            nsdfg_node,
            "neighbor_indices",
        )
        self.state.add_edge(
            nsdfg_node,
            "acc",
            result_node,
            None,
            dace.Memlet(data=result_node.data, subset="0"),
        )

    def _visit_reduce(self, node: gtir.FunCall) -> ValueExpr:
        assert isinstance(node.type, ts.ScalarType)
        op_name, reduce_init, reduce_identity = get_reduce_params(node)
        reduce_wcr = "lambda x, y: " + gtir_python_codegen.format_builtin(op_name, "x", "y")

        result, _ = self.subgraph_builder.add_temp_scalar(self.sdfg, reduce_identity.dc_dtype)
        result_node = self.state.add_access(result)

        input_expr = self.visit(node.args[0])
        assert isinstance(input_expr, (MemletExpr, ValueExpr))
        assert (
            isinstance(input_expr.gt_dtype, ts.ListType)
            and input_expr.gt_dtype.offset_type is not None
        )
        offset_type = input_expr.gt_dtype.offset_type
        offset_provider_type = self.subgraph_builder.get_offset_provider_type(offset_type.value)
        assert isinstance(offset_provider_type, gtx_common.NeighborConnectivityType)

        if offset_provider_type.has_skip_values:
            self._make_reduce_with_skip_values(
                input_expr,
                offset_provider_type,
                reduce_init,
                reduce_identity,
                reduce_wcr,
                result_node,
            )

        else:
            reduce_node = self.state.add_reduce(reduce_wcr, axes=None, identity=reduce_init.value)
            if isinstance(input_expr, MemletExpr):
                self._add_input_data_edge(input_expr.dc_node, input_expr.subset, reduce_node)
            else:
                self.state.add_nedge(
                    input_expr.dc_node,
                    reduce_node,
                    self.sdfg.make_array_memlet(input_expr.dc_node.data),
                )
            self.state.add_nedge(reduce_node, result_node, dace.Memlet(data=result, subset="0"))

        return ValueExpr(result_node, node.type)

    def _split_shift_args(
        self, args: list[gtir.Expr]
    ) -> tuple[tuple[gtir.Expr, gtir.Expr], Optional[list[gtir.Expr]]]:
        """
        Splits the arguments to `shift` builtin function as pairs, each pair containing
        the offset provider and the offset expression in one dimension.
        """
        nargs = len(args)
        assert nargs >= 2 and nargs % 2 == 0
        return (args[-2], args[-1]), args[: nargs - 2] if nargs > 2 else None

    def _visit_shift_multidim(
        self, iterator: gtir.Expr, shift_args: list[gtir.Expr]
    ) -> tuple[gtir.Expr, gtir.Expr, IteratorExpr]:
        """Transforms a multi-dimensional shift into recursive shift calls, each in a single dimension."""
        (offset_provider_arg, offset_value_arg), tail = self._split_shift_args(shift_args)
        if tail:
            node = gtir.FunCall(
                fun=gtir.FunCall(fun=gtir.SymRef(id="shift"), args=tail),
                args=[iterator],
            )
            it = self.visit(node)
        else:
            it = self.visit(iterator)

        assert isinstance(it, IteratorExpr)
        return offset_provider_arg, offset_value_arg, it

    def _make_cartesian_shift(
        self, it: IteratorExpr, offset_dim: gtx_common.Dimension, offset_expr: DataExpr
    ) -> IteratorExpr:
        """Implements cartesian shift along one dimension."""
        assert any(dim == offset_dim for dim, _ in it.field_domain)
        new_index: SymbolExpr | ValueExpr
        index_expr = it.indices[offset_dim]
        if isinstance(index_expr, SymbolExpr) and isinstance(offset_expr, SymbolExpr):
            # purely symbolic expression which can be interpreted at compile time
            new_index = SymbolExpr(
                index_expr.value + offset_expr.value,
                index_expr.dc_dtype,
            )
        else:
            # the offset needs to be calculated by means of a tasklet (i.e. dynamic offset)
            new_index_connector = "shifted_index"
            if isinstance(index_expr, SymbolExpr):
                dynamic_offset_tasklet, connector_mapping = self._add_tasklet(
                    name="dynamic_offset",
                    inputs={"offset"},
                    outputs={new_index_connector},
                    code=f"{new_index_connector} = {index_expr.value} + offset",
                )
            elif isinstance(offset_expr, SymbolExpr):
                dynamic_offset_tasklet, connector_mapping = self._add_tasklet(
                    name="dynamic_offset",
                    inputs={"index"},
                    outputs={new_index_connector},
                    code=f"{new_index_connector} = index + {offset_expr}",
                )
            else:
                dynamic_offset_tasklet, connector_mapping = self._add_tasklet(
                    name="dynamic_offset",
                    inputs={"index", "offset"},
                    outputs={new_index_connector},
                    code=f"{new_index_connector} = index + offset",
                )
            for input_expr, input_connector in [
                (index_expr, connector_mapping.get("index", None)),
                (offset_expr, connector_mapping.get("offset", None)),
            ]:
                if input_connector is None:
                    assert isinstance(input_expr, SymbolExpr)
                elif isinstance(input_expr, MemletExpr):
                    self._add_input_data_edge(
                        input_expr.dc_node,
                        input_expr.subset,
                        dynamic_offset_tasklet,
                        input_connector,
                    )
                elif isinstance(input_expr, ValueExpr):
                    self._add_edge(
                        input_expr.dc_node,
                        None,
                        dynamic_offset_tasklet,
                        input_connector,
                        dace.Memlet(data=input_expr.dc_node.data, subset="0"),
                    )

            if isinstance(index_expr, SymbolExpr):
                dc_dtype = index_expr.dc_dtype
            else:
                dc_dtype = index_expr.dc_node.desc(self.sdfg).dtype

            new_index = self._construct_tasklet_result(
                dc_dtype, dynamic_offset_tasklet, connector_mapping[new_index_connector]
            )

        # a new iterator with a shifted index along one dimension
        shifted_indices = {
            dim: (new_index if dim == offset_dim else index) for dim, index in it.indices.items()
        }
        return IteratorExpr(it.field, it.gt_dtype, it.field_domain, shifted_indices)

    def _make_dynamic_neighbor_offset(
        self,
        offset_expr: MemletExpr | ValueExpr,
        offset_table_node: dace_nodes.AccessNode,
        origin_index: SymbolExpr,
    ) -> ValueExpr:
        """
        Implements access to neighbor connectivity table by means of a tasklet node.

        It requires a dynamic offset value, either obtained from a field/scalar argument (`MemletExpr`)
        or computed by another tasklet (`DataExpr`).
        """
        new_index_connector = "neighbor_index"
        tasklet_node, connector_mapping = self._add_tasklet(
            "dynamic_neighbor_offset",
            {"table", "offset"},
            {new_index_connector},
            f"{new_index_connector} = table[{origin_index.value}, offset]",
        )
        self._add_input_data_edge(
            offset_table_node,
            dace_subsets.Range.from_array(offset_table_node.desc(self.sdfg)),
            tasklet_node,
            connector_mapping["table"],
        )
        if isinstance(offset_expr, MemletExpr):
            self._add_input_data_edge(
                offset_expr.dc_node,
                offset_expr.subset,
                tasklet_node,
                connector_mapping["offset"],
            )
        else:
            self._add_edge(
                offset_expr.dc_node,
                None,
                tasklet_node,
                connector_mapping["offset"],
                dace.Memlet(data=offset_expr.dc_node.data, subset="0"),
            )

        dc_dtype = offset_table_node.desc(self.sdfg).dtype
        return self._construct_tasklet_result(
            dc_dtype, tasklet_node, connector_mapping[new_index_connector]
        )

    def _make_unstructured_shift(
        self,
        it: IteratorExpr,
        conn_type: gtx_common.NeighborConnectivityType,
        conn_node: dace_nodes.AccessNode,
        offset_expr: DataExpr,
    ) -> IteratorExpr:
        """Implements shift in unstructured domain by means of a neighbor table."""
        assert any(dim == conn_type.codomain for dim, _ in it.field_domain)
        neighbor_dim = conn_type.codomain
        origin_dim = conn_type.source_dim
        origin_index = it.indices[origin_dim]
        assert isinstance(origin_index, SymbolExpr)

        shifted_indices = {dim: idx for dim, idx in it.indices.items() if dim != origin_dim}
        if isinstance(offset_expr, SymbolExpr):
            # use memlet to retrieve the neighbor index
            shifted_indices[neighbor_dim] = MemletExpr(
                dc_node=conn_node,
                gt_field=ts.FieldType(
                    dims=[origin_dim],
                    dtype=ts.ListType(
                        element_type=tt.from_dtype(conn_type.dtype), offset_type=_CONST_DIM
                    ),
                ),
                subset=dace_subsets.Range.from_string(f"{origin_index.value}, {offset_expr.value}"),
            )
        else:
            # dynamic offset: we cannot use a memlet to retrieve the offset value, use a tasklet node
            shifted_indices[neighbor_dim] = self._make_dynamic_neighbor_offset(
                offset_expr, conn_node, origin_index
            )

        return IteratorExpr(it.field, it.gt_dtype, it.field_domain, shifted_indices)

    def _visit_shift(self, node: gtir.FunCall) -> IteratorExpr:
        assert isinstance(node.fun, gtir.FunCall)
        # the iterator to be shifted is the node argument, while the shift arguments
        # are provided by the nested function call; the shift arguments consist of
        # the offset provider and the offset value in each dimension to be shifted
        offset_provider_arg, offset_value_arg, it = self._visit_shift_multidim(
            node.args[0], node.fun.args
        )

        # first argument of the shift node is the offset provider
        assert isinstance(offset_provider_arg, gtir.OffsetLiteral)
        offset = offset_provider_arg.value
        assert isinstance(offset, str)
        offset_provider_type = self.subgraph_builder.get_offset_provider_type(offset)
        # second argument should be the offset value, which could be a symbolic expression or a dynamic offset
        offset_expr = (
            SymbolExpr(offset_value_arg.value, gtir_to_sdfg_types.INDEX_DTYPE)
            if isinstance(offset_value_arg, gtir.OffsetLiteral)
            else self.visit(offset_value_arg)
        )

        if isinstance(offset_provider_type, gtx_common.Dimension):
            return self._make_cartesian_shift(it, offset_provider_type, offset_expr)
        else:
            # initially, the storage for the connectivity tables is created as transient;
            # when the tables are used, the storage is changed to non-transient,
            # so the corresponding arrays are supposed to be allocated by the SDFG caller
            offset_table = gtx_dace_args.connectivity_identifier(offset)
            self.sdfg.arrays[offset_table].transient = False
            offset_table_node = self.state.add_access(offset_table)

            return self._make_unstructured_shift(
                it, offset_provider_type, offset_table_node, offset_expr
            )

    def _visit_generic_builtin(self, node: gtir.FunCall) -> ValueExpr:
        """
        Generic handler called by `visit_FunCall()` when it encounters
        a builtin function that does not match any other specific handler.
        """
        node_internals = []
        node_connections: dict[str, MemletExpr | ValueExpr] = {}
        for i, arg in enumerate(node.args):
            arg_expr = self.visit(arg)
            if isinstance(arg_expr, MemletExpr | ValueExpr):
                # the argument value is the result of a tasklet node or direct field access
                connector = f"arg{i}"
                node_connections[connector] = arg_expr
                node_internals.append(connector)
            else:
                assert isinstance(arg_expr, SymbolExpr)
                # use the argument value without adding any connector
                node_internals.append(arg_expr.value)

        assert isinstance(node.fun, gtir.SymRef)
        builtin_name = str(node.fun.id)
        # use tasklet connectors as expression arguments
        code = gtir_python_codegen.format_builtin(builtin_name, *node_internals)

        out_connector = "result"
        tasklet_node, connector_mapping = self._add_tasklet(
            name=builtin_name,
            inputs=set(node_connections.keys()),
            outputs={out_connector},
            code="{} = {}".format(out_connector, code),
        )

        for connector, arg_expr in node_connections.items():
            if isinstance(arg_expr, ValueExpr):
                self._add_edge(
                    arg_expr.dc_node,
                    None,
                    tasklet_node,
                    connector_mapping[connector],
                    dace.Memlet(data=arg_expr.dc_node.data, subset="0"),
                )
            else:
                self._add_input_data_edge(
                    arg_expr.dc_node,
                    arg_expr.subset,
                    tasklet_node,
                    connector_mapping[connector],
                )

        if isinstance(node.type, ts.ListType):
            # The only builtin function (so far) handled here that returns a list
            # is 'make_const_list'. There are other builtin functions (map_, neighbors)
            # that return a list but they are handled in specialized visit methods.
            # This method (the generic visitor for builtin functions) always returns
            # a single value. This is also the case of 'make_const_list' expression:
            # it simply broadcasts a scalar on the local domain of another expression,
            # for example 'map_(plus)(neighbors(V2Eₒ, it), make_const_list(1.0))'.
            # Therefore we handle `ListType` as a single-element array with shape (1,)
            # that will be accessed in a map expression on a local domain.
            assert isinstance(node.type.element_type, ts.ScalarType)
            dc_dtype = gtx_dace_args.as_dace_type(node.type.element_type)
            # In order to ease the lowring of the parent expression on local dimension,
            # we represent the scalar value as a single-element 1D array.
            use_array = True
        else:
            assert isinstance(node.type, ts.ScalarType)
            dc_dtype = gtx_dace_args.as_dace_type(node.type)
            use_array = False

        return self._construct_tasklet_result(
            dc_dtype, tasklet_node, connector_mapping[out_connector], use_array=use_array
        )

    def _visit_make_tuple(self, node: gtir.FunCall) -> tuple[IteratorExpr | DataExpr]:
        assert cpm.is_call_to(node, "make_tuple")
        return tuple(self.visit(arg) for arg in node.args)

    def _visit_tuple_get(
        self, node: gtir.FunCall
    ) -> IteratorExpr | DataExpr | tuple[IteratorExpr | DataExpr]:
        assert cpm.is_call_to(node, "tuple_get")
        assert len(node.args) == 2

        if not isinstance(node.args[0], gtir.Literal):
            raise ValueError("Tuple can only be subscripted with compile-time constants.")
        assert ti.is_integral(node.args[0].type)
        index = int(node.args[0].value)

        tuple_fields = self.visit(node.args[1])
        return tuple_fields[index]

    def visit_FunCall(self, node: gtir.FunCall) -> MaybeNestedInTuple[IteratorExpr | DataExpr]:
        if cpm.is_call_to(node, "deref"):
            return self._visit_deref(node)

        elif cpm.is_call_to(node, "if_"):
            return self._visit_if(node)

        elif cpm.is_call_to(node, "neighbors"):
            return self._visit_neighbors(node)

        elif cpm.is_call_to(node, "list_get"):
            return self._visit_list_get(node)

        elif cpm.is_call_to(node, "make_tuple"):
            return self._visit_make_tuple(node)

        elif cpm.is_call_to(node, "tuple_get"):
            return self._visit_tuple_get(node)

        elif cpm.is_applied_map(node):
            return self._visit_map(node)

        elif cpm.is_applied_reduce(node):
            return self._visit_reduce(node)

        elif cpm.is_applied_shift(node):
            return self._visit_shift(node)

        elif cpm.is_let(node):
            let_node = node.fun
            let_args = [self.visit(arg) for arg in node.args]
            for p, arg in zip(node.fun.params, let_args, strict=True):
                self.symbol_map[str(p.id)] = arg
            return self.visit(let_node.expr)

        elif isinstance(node.fun, gtir.SymRef):
            return self._visit_generic_builtin(node)

        else:
            raise NotImplementedError(f"Invalid 'FunCall' node: {node}.")

    def visit_Lambda(self, node: gtir.Lambda) -> MaybeNestedInTuple[DataflowOutputEdge]:
        def _visit_Lambda_impl(
            output_expr: DataflowOutputEdge | ValueExpr | MemletExpr | SymbolExpr,
        ) -> DataflowOutputEdge:
            if isinstance(output_expr, DataflowOutputEdge):
                return output_expr
            if isinstance(output_expr, ValueExpr):
                return DataflowOutputEdge(self.state, output_expr)

            if isinstance(output_expr, MemletExpr):
                # special case where the field operator is simply copying data from source to destination node
                output_dtype = output_expr.dc_node.desc(self.sdfg).dtype
                tasklet_node, connector_mapping = self._add_tasklet(
                    name="copy", inputs={"inp"}, outputs={"out"}, code="out = inp"
                )
                self._add_input_data_edge(
                    output_expr.dc_node,
                    output_expr.subset,
                    tasklet_node,
                    connector_mapping["inp"],
                )
            else:
                # even simpler case, where a constant value is written to destination node
                output_dtype = output_expr.dc_dtype
                tasklet_node, connector_mapping = self._add_tasklet(
                    name="write", inputs={}, outputs={"out"}, code=f"out = {output_expr.value}"
                )

            output_expr = self._construct_tasklet_result(
                output_dtype, tasklet_node, connector_mapping["out"]
            )
            return DataflowOutputEdge(self.state, output_expr)

        result = self.visit(node.expr)

        return (
            gtx_utils.tree_map(_visit_Lambda_impl)(result)
            if isinstance(result, tuple)
            else _visit_Lambda_impl(result)
        )

    def visit_Literal(self, node: gtir.Literal) -> SymbolExpr:
        dc_dtype = gtx_dace_args.as_dace_type(node.type)
        return SymbolExpr(node.value, dc_dtype)

    def visit_SymRef(self, node: gtir.SymRef) -> MaybeNestedInTuple[IteratorExpr | DataExpr]:
        param = str(node.id)
        if param in self.symbol_map:
            return self.symbol_map[param]
        # if not in the lambda symbol map, this must be a symref to a builtin function
        assert param in gtir_python_codegen.MATH_BUILTINS_MAPPING
        return SymbolExpr(param, dace.string)


def translate_lambda_to_dataflow(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    sdfg_builder: gtir_to_sdfg.DataflowBuilder,
    node: gtir.Lambda,
    args: Sequence[MaybeNestedInTuple[IteratorExpr | MemletExpr | ValueExpr]],
) -> tuple[list[DataflowInputEdge], MaybeNestedInTuple[DataflowOutputEdge]]:
    """
    Entry point to visit a `Lambda` node and lower it to a dataflow graph,
    that can be instantiated inside a map scope implementing the field operator.

    It calls `LambdaToDataflow.visit_let()` to map the lambda arguments to internal
    parameters and visit the let-statements (if any), which always appear as outermost
    nodes. Finally, the visitor returns the output edge of the dataflow.

    Args:
        sdfg: The SDFG where the dataflow graph will be instantiated.
        state: The SDFG state where the dataflow graph will be instantiated.
        sdfg_builder: Helper class to build the dataflow inside the given SDFG.
        node: Lambda node to visit.
        args: Arguments passed to lambda node.

    Returns:
        A tuple of two elements:
        - List of connections for data inputs to the dataflow.
        - Tree representation of output data connections.
    """
    taskgen = LambdaToDataflow(sdfg, state, sdfg_builder)
    for p, arg in zip(node.params, args, strict=True):
        taskgen.symbol_map[str(p.id)] = arg

    lambda_output = taskgen.visit(node)

    # remove access nodes to lambda symbols which were not used
    flat_arg_nodes = (
        x.field if isinstance(x, IteratorExpr) else x.dc_node  # type: ignore[attr-defined]
        for x in gtx_utils.flatten_nested_tuple(tuple(args))
    )
    state.remove_nodes_from([node for node in flat_arg_nodes if state.degree(node) == 0])

    return taskgen.input_edges, lambda_output
